from __future__ import annotations

import logging
import textwrap
from collections import OrderedDict, defaultdict
from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Optional, Union

import rflx.typing_ as rty
from rflx import expression as expr, lang, model
from rflx.common import STDIN, unique
from rflx.error import Location, RecordFluxError, Severity, Subsystem, fail
from rflx.identifier import ID, StrID
from rflx.integration import Integration
from rflx.model import AlwaysVerify, Cache, declaration as decl, statement as stmt
from rflx.specification.const import RESERVED_WORDS

from . import style

log = logging.getLogger(__name__)


def node_location(node: lang.RFLXNode, filename: Path, end_location: bool = False) -> Location:
    assert node.token_start
    assert node.token_end
    start = node.token_start.sloc_range
    end = node.token_end.sloc_range
    return Location(
        start=(
            (start.start.line, start.start.column)
            if not end_location
            else (end.end.line, end.end.column)
        ),
        source=filename,
        end=(end.end.line, end.end.column),
    )


def type_location(identifier: ID, node: lang.RFLXNode) -> Location:
    """
    Create a location object.

    The location object covers the area from the start of an identifier to the end of a node.
    """
    assert identifier.location is not None
    assert identifier.location.source is not None
    return Location(
        identifier.location.start,
        identifier.location.source,
        node_location(node, identifier.location.source).end,
    )


def diagnostics_to_error(
    diagnostics: Sequence[lang.Diagnostic],
    error: RecordFluxError,
    filename: Path,
) -> bool:
    """Append langkit diagnostics to RecordFlux error. Return True if error occured."""

    if len(diagnostics) == 0:
        return False

    for diag in diagnostics:
        loc = diag.sloc_range
        error.extend(
            [
                (
                    diag.message,
                    Subsystem.PARSER,
                    Severity.ERROR,
                    Location(
                        start=(loc.start.line, loc.start.column),
                        source=filename,
                        end=(loc.end.line, loc.end.column),
                    ),
                ),
            ],
        )
    return True


def readable_name(name: str) -> str:
    """Convert entity name to a readable format, e.g. CaseExpression to 'case expression'."""

    words = []
    start_index = 0
    for i in range(1, len(name)):
        if name[i].isupper():
            words.append(name[start_index:i].lower())
            start_index = i
    words.append(name[start_index:].lower())
    return " ".join(words)


def validate_handler(
    error: RecordFluxError,
    context: str,
    entity: lang.RFLXNode,
    handlers: list[str],
    filename: Path,
) -> None:
    if entity.kind_name not in handlers:
        error.extend(
            [
                (
                    f"{readable_name(entity.kind_name)} unsupported in {context} context",
                    Subsystem.PARSER,
                    Severity.ERROR,
                    node_location(entity, filename),
                ),
            ],
        )
        error.propagate()


def create_description(description: Optional[lang.Description] = None) -> Optional[str]:
    if description:
        assert isinstance(description.text, str)
        return description.text.split('"')[1]
    return None


def create_transition(
    error: RecordFluxError,
    transition: lang.Transition,
    filename: Path,
) -> model.Transition:
    if transition.kind_name not in ("Transition", "ConditionalTransition"):
        raise NotImplementedError(f"Transition kind {transition.kind_name} unsupported")
    target = create_id_or_null(error, transition.f_target, filename)
    condition: expr.Expr = expr.TRUE
    description = create_description(transition.f_description)
    if isinstance(transition, lang.ConditionalTransition):
        condition = create_bool_expression(error, transition.f_condition, filename)
    return model.Transition(target, condition, description, node_location(transition, filename))


def create_reset(error: RecordFluxError, reset: lang.Statement, filename: Path) -> stmt.Statement:
    assert isinstance(reset, lang.Reset)
    return stmt.Reset(
        create_id(error, reset.f_identifier, filename),
        {
            create_id(error, c.f_identifier, filename): create_expression(
                error,
                c.f_expression,
                filename,
            )
            for c in reset.f_associations
        },
        location=node_location(reset, filename),
    )


def create_assignment(
    error: RecordFluxError,
    assignment: lang.Statement,
    filename: Path,
) -> stmt.Statement:
    assert isinstance(assignment, lang.Assignment)
    return stmt.VariableAssignment(
        create_id(error, assignment.f_identifier, filename),
        create_expression(error, assignment.f_expression, filename),
        location=node_location(assignment, filename),
    )


def create_message_field_assignment(
    error: RecordFluxError,
    assignment: lang.Statement,
    filename: Path,
) -> stmt.Statement:
    assert isinstance(assignment, lang.MessageFieldAssignment)
    return stmt.MessageFieldAssignment(
        create_id(error, assignment.f_message, filename),
        create_id(error, assignment.f_field, filename),
        create_expression(error, assignment.f_expression, filename),
        location=node_location(assignment, filename),
    )


def create_attribute_statement(
    error: RecordFluxError,
    expression: lang.Statement,
    filename: Path,
) -> stmt.Statement:
    assert isinstance(expression, lang.AttributeStatement)
    attrs = {
        "Append": stmt.Append,
        "Extend": stmt.Extend,
        "Read": stmt.Read,
        "Write": stmt.Write,
    }

    constructor = attrs[expression.f_attr.text]

    return constructor(
        create_id(error, expression.f_identifier, filename),
        create_expression(error, expression.f_expression, filename),
        location=node_location(expression, filename),
    )


def create_statement(
    error: RecordFluxError,
    statement: lang.Statement,
    filename: Path,
) -> stmt.Statement:
    handlers = {
        "Reset": create_reset,
        "Assignment": create_assignment,
        "MessageFieldAssignment": create_message_field_assignment,
        "AttributeStatement": create_attribute_statement,
    }
    validate_handler(error, "statement", statement, list(handlers.keys()), filename)
    return handlers[statement.kind_name](error, statement, filename)


def create_state(
    error: RecordFluxError,
    state: lang.State,
    package: ID,
    filename: Path,
) -> model.State:
    location = node_location(state, filename)
    identifier = create_id(error, state.f_identifier, filename)
    assert isinstance(state.f_body, lang.StateBody)
    if state.f_identifier.text != state.f_body.f_end_identifier.text:
        error.extend(
            [
                (
                    "inconsistent state identifier: "
                    f"{state.f_identifier.text} /= {state.f_body.f_end_identifier.text}",
                    Subsystem.PARSER,
                    Severity.ERROR,
                    location,
                ),
            ],
        )
    transitions = [
        create_transition(error, t, filename) for t in state.f_body.f_conditional_transitions
    ]
    transitions.append(create_transition(error, state.f_body.f_final_transition, filename))
    exception_transition = (
        create_transition(error, state.f_body.f_exception_transition, filename)
        if state.f_body.f_exception_transition
        else None
    )
    actions = [create_statement(error, a, filename) for a in state.f_body.f_actions]
    declarations = [
        create_declaration(error, d, package, filename) for d in state.f_body.f_declarations
    ]
    description = create_description(state.f_description)
    return model.State(
        identifier=identifier,
        transitions=transitions,
        exception_transition=exception_transition,
        actions=actions,
        declarations=declarations,
        description=description,
        location=node_location(state, filename),
    )


def _check_session_identifier(
    error: RecordFluxError,
    session: lang.SessionDecl,
    filename: Path,
) -> None:
    if session.f_identifier.text != session.f_end_identifier.text:
        error.extend(
            [
                (
                    "inconsistent session identifier: "
                    f"{session.f_identifier.text} /= {session.f_end_identifier.text}",
                    Subsystem.PARSER,
                    Severity.ERROR,
                    node_location(session, filename),
                ),
            ],
        )


def create_session(
    error: RecordFluxError,
    session: lang.SessionDecl,
    package: ID,
    filename: Path,
) -> model.UncheckedSession:
    _check_session_identifier(error, session, filename)

    return model.UncheckedSession(
        package * create_id(error, session.f_identifier, filename),
        [create_state(error, s, package, filename) for s in session.f_states],
        [create_declaration(error, d, package, filename) for d in session.f_declarations],
        [create_formal_declaration(error, p, package, filename) for p in session.f_parameters],
        node_location(session, filename),
    )


def create_id(error: RecordFluxError, identifier: lang.AbstractID, filename: Path) -> ID:
    if isinstance(identifier, lang.UnqualifiedID):
        if identifier.text.lower() in RESERVED_WORDS:
            error.extend(
                [
                    (
                        f'reserved word "{identifier.text}" used as identifier',
                        Subsystem.PARSER,
                        Severity.ERROR,
                        node_location(identifier, filename),
                    ),
                ],
            )
        if identifier.text.upper().startswith("RFLX_"):
            error.extend(
                [
                    (
                        f'illegal identifier "{identifier.text}"',
                        Subsystem.PARSER,
                        Severity.ERROR,
                        node_location(identifier, filename),
                    ),
                    (
                        'identifiers starting with "RFLX_" are reserved for internal use',
                        Subsystem.PARSER,
                        Severity.INFO,
                        node_location(identifier, filename),
                    ),
                ],
            )
        return ID(identifier.text, location=node_location(identifier, filename))
    if isinstance(identifier, lang.ID):
        name = ID(identifier.f_name.text, location=node_location(identifier.f_name, filename))
        if identifier.f_package:
            return (
                ID(
                    identifier.f_package.text,
                    location=node_location(identifier.f_package, filename),
                )
                * name
            )
        return name

    raise NotImplementedError(f"Invalid ID: {identifier.text} {type(identifier)}")


def create_id_or_null(error: RecordFluxError, identifier: lang.AbstractID, filename: Path) -> ID:
    if isinstance(identifier, lang.UnqualifiedID) and identifier.text.lower() == "null":
        return ID("null", location=node_location(identifier, filename))
    return create_id(error, identifier, filename)


def create_sequence(
    error: RecordFluxError,
    identifier: ID,
    _parameters: lang.Parameters,
    sequence: lang.TypeDef,
    filename: Path,
) -> Optional[model.UncheckedSequence]:
    assert isinstance(sequence, lang.SequenceTypeDef)
    element_identifier = model.internal_type_identifier(
        create_id(error, sequence.f_element_type, filename),
        identifier.parent,
    )
    return model.UncheckedSequence(
        identifier,
        element_identifier,
        type_location(identifier, sequence),
    )


def create_numeric_literal(
    _error: RecordFluxError,
    expression: lang.Expr,
    filename: Path,
) -> expr.Expr:
    location = node_location(expression, filename)
    num = expression.text.split("#")
    if len(num) == 1:
        return expr.Number(int(num[0]), location=location)
    if len(num) == 3:
        base = int(num[0])
        return expr.Number(int(num[1], base), base=base, location=location)
    raise NotImplementedError(f"Invalid numeric literal: {expression.text}")


OPERATIONS: Mapping[str, type[expr.BinExpr]] = {
    "OpIn": expr.In,
    "OpNotin": expr.NotIn,
    "OpEq": expr.Equal,
    "OpNeq": expr.NotEqual,
}


def create_binop(error: RecordFluxError, expression: lang.Expr, filename: Path) -> expr.Expr:
    assert isinstance(expression, lang.BinOp)
    loc = node_location(expression, filename)
    if expression.f_op.kind_name in OPERATIONS:
        return OPERATIONS[expression.f_op.kind_name](
            create_expression(error, expression.f_left, filename),
            create_expression(error, expression.f_right, filename),
            location=loc,
        )
    if expression.f_op.kind_name in BOOLEAN_OPERATIONS:
        return BOOLEAN_OPERATIONS[expression.f_op.kind_name](
            create_expression(error, expression.f_left, filename),
            create_expression(error, expression.f_right, filename),
            location=loc,
        )

    left = create_math_expression(error, expression.f_left, filename)
    right = create_math_expression(error, expression.f_right, filename)
    if expression.f_op.kind_name in MATH_OPERATIONS:
        return MATH_OPERATIONS[expression.f_op.kind_name](left, right, location=loc)
    if expression.f_op.kind_name in MATH_COMPARISONS:
        return MATH_COMPARISONS[expression.f_op.kind_name](left, right, location=loc)

    raise NotImplementedError(f"Invalid BinOp {expression.f_op.kind_name} => {expression.text}")


MATH_OPERATIONS: Mapping[str, type[Union[expr.BinExpr, expr.AssExpr]]] = {
    "OpPow": expr.Pow,
    "OpAdd": expr.Add,
    "OpSub": expr.Sub,
    "OpMul": expr.Mul,
    "OpDiv": expr.Div,
    "OpMod": expr.Mod,
}


def create_math_binop(error: RecordFluxError, expression: lang.Expr, filename: Path) -> expr.Expr:
    assert isinstance(expression, lang.BinOp)

    if expression.f_left is None or expression.f_right is None:
        # TODO(eng/recordflux/RecordFlux#1526): This should be rejected by the parser already.
        fail(
            "empty subexpression",
            Subsystem.PARSER,
            location=node_location(expression, filename),
        )
    if expression.f_op.kind_name in MATH_OPERATIONS:
        return MATH_OPERATIONS[expression.f_op.kind_name](
            create_math_expression(error, expression.f_left, filename),
            create_math_expression(error, expression.f_right, filename),
            location=node_location(expression, filename),
        )

    if expression.f_op.kind_name in [*OPERATIONS, *MATH_COMPARISONS, *BOOLEAN_OPERATIONS]:
        fail(
            "boolean expression in math context",
            Subsystem.PARSER,
            location=node_location(expression, filename),
        )

    raise NotImplementedError(
        f"Invalid math BinOp {expression.f_op.kind_name} => {expression.text}",
    )


MATH_COMPARISONS: Mapping[str, type[expr.Relation]] = {
    "OpLt": expr.Less,
    "OpGt": expr.Greater,
    "OpLe": expr.LessEqual,
    "OpGe": expr.GreaterEqual,
}

BOOLEAN_OPERATIONS = {
    "OpAnd": expr.And,
    "OpOr": expr.Or,
}


def create_bool_binop(error: RecordFluxError, expression: lang.Expr, filename: Path) -> expr.Expr:
    assert isinstance(expression, lang.BinOp)

    if expression.f_left is None or expression.f_right is None:
        # TODO(eng/recordflux/RecordFlux#1526): This should be rejected by the parser already.
        fail(
            "empty subexpression",
            Subsystem.PARSER,
            location=node_location(expression, filename),
        )

    if expression.f_op.kind_name in MATH_COMPARISONS:
        return MATH_COMPARISONS[expression.f_op.kind_name](
            create_math_expression(error, expression.f_left, filename),
            create_math_expression(error, expression.f_right, filename),
            location=node_location(expression, filename),
        )
    if expression.f_op.kind_name in BOOLEAN_OPERATIONS:
        return BOOLEAN_OPERATIONS[expression.f_op.kind_name](
            create_bool_expression(error, expression.f_left, filename),
            create_bool_expression(error, expression.f_right, filename),
            location=node_location(expression, filename),
        )
    if expression.f_op.kind_name in OPERATIONS:
        return OPERATIONS[expression.f_op.kind_name](
            create_expression(error, expression.f_left, filename),
            create_expression(error, expression.f_right, filename),
            location=node_location(expression, filename),
        )

    if expression.f_op.kind_name in MATH_OPERATIONS:
        fail(
            "math expression in boolean context",
            Subsystem.PARSER,
            location=node_location(expression, filename),
        )

    raise NotImplementedError(
        f"Invalid bool BinOp {expression.f_op.kind_name} => {expression.text}",
    )


def create_paren_bool_expression(
    error: RecordFluxError,
    expression: lang.Expr,
    filename: Path,
) -> expr.Expr:
    assert isinstance(expression, lang.ParenExpression)
    return create_bool_expression(error, expression.f_data, filename)


def create_paren_math_expression(
    error: RecordFluxError,
    expression: lang.Expr,
    filename: Path,
) -> expr.Expr:
    assert isinstance(expression, lang.ParenExpression)
    return create_math_expression(error, expression.f_data, filename)


def create_paren_expression(
    error: RecordFluxError,
    expression: lang.Expr,
    filename: Path,
) -> expr.Expr:
    assert isinstance(expression, lang.ParenExpression)
    return create_expression(error, expression.f_data, filename)


def create_variable(error: RecordFluxError, expression: lang.Expr, filename: Path) -> expr.Expr:
    assert isinstance(expression, lang.Variable)
    location = node_location(expression, filename)
    if expression.f_identifier.text.lower() in ("true", "false"):
        return expr.Literal(
            create_id(error, expression.f_identifier, filename),
            location=location,
            type_=rty.BOOLEAN,
        )
    return expr.Variable(create_id(error, expression.f_identifier, filename), location=location)


def create_math_attribute(
    error: RecordFluxError,
    expression: lang.Expr,
    filename: Path,
) -> expr.Expr:
    assert isinstance(expression, lang.Attribute)
    inner = create_expression(error, expression.f_expression, filename)
    if expression.f_kind.kind_name == "AttrLast":
        return expr.Last(inner)
    if expression.f_kind.kind_name == "AttrFirst":
        return expr.First(inner)
    if expression.f_kind.kind_name == "AttrSize":
        return expr.Size(inner)
    raise NotImplementedError(
        f"Invalid math attribute: {expression.f_kind.kind_name} => {expression.text}",
    )


def create_attribute(error: RecordFluxError, expression: lang.Expr, filename: Path) -> expr.Expr:
    assert isinstance(expression, lang.Attribute)
    inner = create_expression(error, expression.f_expression, filename)
    if expression.f_kind.kind_name == "AttrLast":
        return expr.Last(inner)
    if expression.f_kind.kind_name == "AttrFirst":
        return expr.First(inner)
    if expression.f_kind.kind_name == "AttrSize":
        return expr.Size(inner)
    if expression.f_kind.kind_name == "AttrValidChecksum":
        return expr.ValidChecksum(inner)
    if expression.f_kind.kind_name == "AttrHasData":
        return expr.HasData(inner)
    if expression.f_kind.kind_name == "AttrHead":
        return expr.Head(inner)
    if expression.f_kind.kind_name == "AttrOpaque":
        return expr.Opaque(inner)
    if expression.f_kind.kind_name == "AttrPresent":
        return expr.Present(inner)
    if expression.f_kind.kind_name == "AttrValid":
        return expr.Valid(inner)
    raise NotImplementedError(
        f"Invalid attribute: {expression.f_kind.kind_name} => {expression.text}",
    )


def create_sequence_aggregate(
    error: RecordFluxError,
    expression: lang.Expr,
    filename: Path,
) -> expr.Expr:
    assert isinstance(expression, lang.SequenceAggregate)
    return expr.Aggregate(
        *[create_math_expression(error, v, filename) for v in expression.f_values],
        location=node_location(expression, filename),
    )


def create_string_literal(
    _error: RecordFluxError,
    expression: lang.Expr,
    filename: Path,
) -> expr.Expr:
    return expr.String(
        expression.text.split('"')[1],
        location=node_location(expression, filename),
    )


def create_call(error: RecordFluxError, expression: lang.Expr, filename: Path) -> expr.Expr:
    assert isinstance(expression, lang.Call)
    return expr.Call(
        create_id(error, expression.f_identifier, filename),
        [create_expression(error, a, filename) for a in expression.f_arguments],
        location=node_location(expression, filename),
    )


def create_quantified_expression(
    error: RecordFluxError,
    expression: lang.Expr,
    filename: Path,
) -> expr.Expr:
    assert isinstance(expression, lang.QuantifiedExpression)
    param_id = create_id(error, expression.f_parameter_identifier, filename)
    iterable = create_expression(error, expression.f_iterable, filename)
    predicate = create_expression(error, expression.f_predicate, filename)
    location = node_location(expression, filename)
    if expression.f_operation.kind_name == "QuantifierAll":
        return expr.ForAllIn(param_id, iterable, predicate, location)
    if expression.f_operation.kind_name == "QuantifierSome":
        return expr.ForSomeIn(param_id, iterable, predicate, location)

    raise NotImplementedError(f"Invalid quantified: {expression.f_operation.text}")


def create_binding(error: RecordFluxError, expression: lang.Expr, filename: Path) -> expr.Expr:
    assert isinstance(expression, lang.Binding)
    error.extend(
        [
            (
                "bindings are not supported",
                Subsystem.PARSER,
                Severity.ERROR,
                node_location(expression, filename),
            ),
        ],
    )
    return create_expression(error, expression.f_expression, filename)


def create_variable_decl(
    error: RecordFluxError,
    declaration: lang.LocalDecl,
    package: ID,
    filename: Path,
) -> decl.BasicDeclaration:
    assert isinstance(declaration, lang.VariableDecl)
    initializer = (
        create_expression(error, declaration.f_initializer, filename)
        if declaration.f_initializer
        else None
    )
    return decl.VariableDeclaration(
        create_id(error, declaration.f_identifier, filename),
        model.internal_type_identifier(
            create_id(error, declaration.f_type_identifier, filename),
            package,
        ),
        initializer,
        location=node_location(declaration, filename),
    )


def create_channel_decl(
    error: RecordFluxError,
    declaration: lang.FormalDecl,
    _package: ID,
    filename: Path,
) -> decl.FormalDeclaration:
    assert isinstance(declaration, lang.FormalChannelDecl)
    readable = False
    writable = False

    grouped = defaultdict(list)
    for p in declaration.f_parameters:
        grouped[p.kind_name].append(node_location(p, filename))

    for name, locations in grouped.items():
        check_duplicate_aspect(error, name, locations)
        if name == "Readable":
            readable = True
        elif name == "Writable":
            writable = True
        else:
            raise NotImplementedError(f"channel parameter: {name}")
    return decl.ChannelDeclaration(
        create_id(error, declaration.f_identifier, filename),
        readable=readable,
        writable=writable,
        location=node_location(declaration, filename),
    )


def create_renaming_decl(
    error: RecordFluxError,
    declaration: lang.LocalDecl,
    package: ID,
    filename: Path,
) -> decl.BasicDeclaration:
    assert isinstance(declaration, lang.RenamingDecl)
    selected = create_expression(error, declaration.f_expression, filename)
    assert isinstance(selected, expr.Selected)
    return decl.RenamingDeclaration(
        create_id(error, declaration.f_identifier, filename),
        model.internal_type_identifier(
            create_id(error, declaration.f_type_identifier, filename),
            package,
        ),
        selected,
        location=node_location(declaration, filename),
    )


def create_function_decl(
    error: RecordFluxError,
    declaration: lang.FormalDecl,
    package: ID,
    filename: Path,
) -> decl.FormalDeclaration:
    assert isinstance(declaration, lang.FormalFunctionDecl)
    arguments = []
    if declaration.f_parameters:
        arguments = [
            decl.Argument(
                create_id(error, p.f_identifier, filename),
                model.internal_type_identifier(
                    create_id(error, p.f_type_identifier, filename),
                    package,
                ),
            )
            for p in declaration.f_parameters.f_parameters
        ]
    return decl.FunctionDeclaration(
        create_id(error, declaration.f_identifier, filename),
        arguments,
        model.internal_type_identifier(
            create_id(error, declaration.f_return_type_identifier, filename),
            package,
        ),
        location=node_location(declaration, filename),
    )


def create_negation(error: RecordFluxError, expression: lang.Expr, filename: Path) -> expr.Expr:
    assert isinstance(expression, lang.Negation)
    math_expr = create_math_expression(error, expression.f_data, filename)
    assert isinstance(math_expr, expr.Number)
    return expr.Number(-math_expr.value, math_expr.base, node_location(expression, filename))


def create_concatenation(
    error: RecordFluxError,
    expression: lang.Expr,
    filename: Path,
) -> expr.Expr:
    assert isinstance(expression, lang.Concatenation)
    left = create_expression(error, expression.f_left, filename)
    right = create_expression(error, expression.f_right, filename)
    assert isinstance(left, expr.Aggregate)
    assert isinstance(right, expr.Aggregate)
    return expr.Aggregate(
        *(left.elements + right.elements),
        location=node_location(expression, filename),
    )


def create_comprehension(
    error: RecordFluxError,
    expression: lang.Expr,
    filename: Path,
) -> expr.Expr:
    assert isinstance(expression, lang.Comprehension)
    condition = (
        create_bool_expression(error, expression.f_condition, filename)
        if expression.f_condition
        else expr.TRUE
    )
    return expr.Comprehension(
        create_id(error, expression.f_iterator, filename),
        create_expression(error, expression.f_sequence, filename),
        create_expression(error, expression.f_selector, filename),
        condition,
        node_location(expression, filename),
    )


def create_selected(error: RecordFluxError, expression: lang.Expr, filename: Path) -> expr.Expr:
    assert isinstance(expression, lang.SelectNode)
    return expr.Selected(
        create_expression(error, expression.f_expression, filename),
        create_id(error, expression.f_selector, filename),
        location=node_location(expression, filename),
    )


def create_case(error: RecordFluxError, expression: lang.Expr, filename: Path) -> expr.Expr:
    assert isinstance(expression, lang.CaseExpression)

    def create_choice(
        value: Union[lang.AbstractID, lang.Expr],
        filename: Path,
    ) -> Union[ID, expr.Number]:
        if isinstance(value, lang.AbstractID):
            return create_id(error, value, filename)
        assert isinstance(value, lang.Expr)
        result = create_numeric_literal(error, value, filename)
        assert isinstance(result, expr.Number)
        return result

    choices: Sequence[tuple[Sequence[Union[ID, expr.Number]], expr.Expr]] = [
        (
            [
                create_choice(s, filename)
                for s in c.f_selectors
                if isinstance(s, (lang.AbstractID, lang.Expr))
            ],
            create_expression(error, c.f_expression, filename),
        )
        for c in expression.f_choices
    ]

    return expr.CaseExpr(
        create_expression(error, expression.f_expression, filename),
        choices,
        location=node_location(expression, filename),
    )


def create_conversion(error: RecordFluxError, expression: lang.Expr, filename: Path) -> expr.Expr:
    assert isinstance(expression, lang.Conversion)
    return expr.Conversion(
        create_id(error, expression.f_target_identifier, filename),
        create_expression(error, expression.f_argument, filename),
        location=node_location(expression, filename),
    )


def create_message_aggregate(
    error: RecordFluxError,
    expression: lang.Expr,
    filename: Path,
) -> expr.Expr:
    assert isinstance(expression, lang.MessageAggregate)
    values: Mapping[StrID, expr.Expr] = {}
    if isinstance(expression.f_values, lang.NullMessageAggregate):
        values = {}
    elif isinstance(expression.f_values, lang.MessageAggregateAssociations):
        values = {
            create_id(error, c.f_identifier, filename): create_expression(
                error,
                c.f_expression,
                filename,
            )
            for c in expression.f_values.f_associations
        }
    else:
        raise NotImplementedError(f"invalid message field: {type(expression.f_values)}")

    return expr.MessageAggregate(
        create_id(error, expression.f_identifier, filename),
        values,
        location=node_location(expression, filename),
    )


EXPRESSION_MAP = {
    "NumericLiteral": create_numeric_literal,
    "ParenExpression": create_paren_expression,
    "Variable": create_variable,
    "Attribute": create_attribute,
    "SequenceAggregate": create_sequence_aggregate,
    "StringLiteral": create_string_literal,
    "Call": create_call,
    "QuantifiedExpression": create_quantified_expression,
    "Binding": create_binding,
    "Negation": create_negation,
    "Concatenation": create_concatenation,
    "Comprehension": create_comprehension,
    "SelectNode": create_selected,
    "Conversion": create_conversion,
    "MessageAggregate": create_message_aggregate,
    "BinOp": create_binop,
    "CaseExpression": create_case,
}


def create_expression(error: RecordFluxError, expression: lang.Expr, filename: Path) -> expr.Expr:
    validate_handler(error, "expression", expression, list(EXPRESSION_MAP.keys()), filename)
    return EXPRESSION_MAP[expression.kind_name](error, expression, filename)


def create_declaration(
    error: RecordFluxError,
    declaration: lang.LocalDecl,
    package: ID,
    filename: Path,
) -> decl.BasicDeclaration:
    handlers = {
        "VariableDecl": create_variable_decl,
        "RenamingDecl": create_renaming_decl,
    }
    validate_handler(error, "declaration", declaration, list(handlers.keys()), filename)
    return handlers[declaration.kind_name](error, declaration, package, filename)


def create_formal_declaration(
    error: RecordFluxError,
    declaration: lang.FormalDecl,
    package: ID,
    filename: Path,
) -> decl.FormalDeclaration:
    handlers = {
        "FormalChannelDecl": create_channel_decl,
        "FormalFunctionDecl": create_function_decl,
    }
    validate_handler(error, "formal declaration", declaration, list(handlers.keys()), filename)
    return handlers[declaration.kind_name](error, declaration, package, filename)


def create_math_expression(
    error: RecordFluxError,
    expression: lang.Expr,
    filename: Path,
) -> expr.Expr:
    handlers = {
        "NumericLiteral": create_numeric_literal,
        "BinOp": create_math_binop,
        "ParenExpression": create_paren_math_expression,
        "Variable": create_variable,
        "Call": create_call,
        "Binding": create_binding,
        "Negation": create_negation,
        "Attribute": create_math_attribute,
        "SelectNode": create_selected,
        "SequenceAggregate": create_sequence_aggregate,
    }
    validate_handler(error, "math expression", expression, list(handlers.keys()), filename)
    result = handlers[expression.kind_name](error, expression, filename)
    if result.type_ == rty.BOOLEAN:
        fail(
            "boolean expression in math context",
            Subsystem.PARSER,
            location=node_location(expression, filename),
        )
    return result


def create_bool_expression(
    error: RecordFluxError,
    expression: lang.Expr,
    filename: Path,
) -> expr.Expr:
    handlers = {
        "BinOp": create_bool_binop,
        "ParenExpression": create_paren_bool_expression,
        "Attribute": create_attribute,
        "Call": create_call,
        "Variable": create_variable,
        "QuantifiedExpression": create_quantified_expression,
        "Binding": create_binding,
        "SelectNode": create_selected,
        "CaseExpression": create_case,
    }
    if expression.kind_name == "NumericLiteral":
        fail(
            "math expression in boolean context",
            Subsystem.PARSER,
            location=node_location(expression, filename),
        )
    validate_handler(error, "boolean expression", expression, list(handlers.keys()), filename)
    return handlers[expression.kind_name](error, expression, filename)


def create_modular(
    error: RecordFluxError,
    identifier: ID,
    _parameters: lang.Parameters,
    modular: lang.TypeDef,
    filename: Path,
) -> None:
    assert isinstance(modular, lang.ModularTypeDef)
    modulus = create_math_expression(error, modular.f_mod, filename).simplified()
    assert isinstance(modulus, expr.Number)
    upper = int(modulus) - 1
    error.extend(
        [
            (
                "modular integer types are not supported",
                Subsystem.PARSER,
                Severity.ERROR,
                type_location(identifier, modular),
            ),
            (
                f'use "type {identifier.name} is range 0 .. {upper}'
                f' with Size => {upper.bit_length()}" instead',
                Subsystem.PARSER,
                Severity.INFO,
                type_location(identifier, modular),
            ),
        ],
    )


def create_range(
    error: RecordFluxError,
    identifier: ID,
    _parameters: lang.Parameters,
    rangetype: lang.TypeDef,
    filename: Path,
) -> Optional[model.UncheckedInteger]:
    assert isinstance(rangetype, lang.RangeTypeDef)
    if rangetype.f_size.f_identifier.text != "Size":
        error.extend(
            [
                (
                    f'invalid aspect "{rangetype.f_size.f_identifier.text}" '
                    f'for range type "{identifier}"',
                    Subsystem.PARSER,
                    Severity.ERROR,
                    node_location(rangetype, filename),
                ),
            ],
        )
        return None
    size = create_math_expression(error, rangetype.f_size.f_value, filename)
    return model.UncheckedInteger(
        identifier,
        create_math_expression(error, rangetype.f_first, filename),
        create_math_expression(error, rangetype.f_last, filename),
        size,
        type_location(identifier, rangetype),
    )


def create_null_message(
    _error: RecordFluxError,
    identifier: ID,
    _parameters: lang.Parameters,
    message: lang.TypeDef,
    _filename: Path,
) -> Optional[model.UncheckedMessage]:
    assert isinstance(message, lang.NullMessageTypeDef)
    return model.UncheckedMessage(
        identifier,
        [],
        [],
        [],
        {},
        {},
        location=type_location(identifier, message),
    )


def create_message(
    error: RecordFluxError,
    identifier: ID,
    parameters: lang.Parameters,
    message: lang.TypeDef,
    filename: Path,
) -> Optional[model.UncheckedMessage]:
    assert isinstance(message, lang.MessageTypeDef)
    fields = message.f_message_fields

    def get_parameters(param: lang.Parameters) -> Optional[lang.ParameterList]:
        if not param:
            return None
        assert isinstance(param.f_parameters, lang.ParameterList)
        return param.f_parameters

    parameter_types = create_message_field_types(
        error,
        identifier,
        (
            (parameter.f_identifier, parameter.f_type_identifier, [])
            for parameter in get_parameters(parameters) or []
        ),
        filename,
    )
    field_types = create_message_field_types(
        error,
        identifier,
        (
            (field.f_identifier, field.f_type_identifier, field.f_type_arguments)
            for field in fields.f_fields
        ),
        filename,
    )
    structure = create_message_structure(error, fields, filename)
    checksum_aspects, byte_order_aspect = parse_aspects(error, message.f_aspects, filename)

    return model.UncheckedMessage(
        identifier,
        structure,
        parameter_types,
        field_types,
        checksum_aspects,
        byte_order_aspect,
        type_location(identifier, message),
    )


def create_message_field_types(
    error: RecordFluxError,
    identifier: ID,
    fields: Iterable[tuple[lang.AbstractID, lang.AbstractID, object]],
    filename: Path,
) -> list[tuple[model.Field, ID, list[tuple[ID, expr.Expr]]]]:
    result: list[tuple[model.Field, ID, list[tuple[ID, expr.Expr]]]] = []
    for field_identifier, type_identifier, type_arguments in fields:
        qualified_type_identifier = model.internal_type_identifier(
            create_id(error, type_identifier, filename),
            identifier.parent,
        )
        field = model.Field(create_id(error, field_identifier, filename))
        result.append(
            (
                field,
                qualified_type_identifier,
                (
                    create_message_arguments(error, type_arguments, filename)
                    if isinstance(type_arguments, lang.TypeArgumentList)
                    else []
                ),
            ),
        )
    return result


def create_message_arguments(
    error: RecordFluxError,
    type_arguments: lang.TypeArgumentList,
    filename: Path,
) -> list[tuple[ID, expr.Expr]]:
    return [
        (
            create_id(error, arg.f_identifier, filename),
            create_expression(error, arg.f_expression, filename),
        )
        for arg in type_arguments
    ]


def create_message_structure(
    error: RecordFluxError,
    fields: lang.MessageFields,
    filename: Path,
) -> list[model.Link]:
    def extract_aspect(aspects: lang.AspectList) -> tuple[expr.Expr, expr.Expr]:
        size: expr.Expr = expr.UNDEFINED
        first: expr.Expr = expr.UNDEFINED

        grouped = defaultdict(list)
        for aspect in aspects:
            grouped[aspect.f_identifier.text].append(
                (aspect, node_location(aspect.f_identifier, filename)),
            )

        for name, locations in grouped.items():
            check_duplicate_aspect(error, name, [l for _, l in locations])
            aspect, location = locations[0]

            if name == "Size":
                size = create_math_expression(error, aspect.f_value, filename)
            elif name == "First":
                first = create_math_expression(error, aspect.f_value, filename)
            else:
                error.extend(
                    [
                        (
                            f'invalid aspect "{name}"',
                            Subsystem.PARSER,
                            Severity.ERROR,
                            location,
                        ),
                    ],
                )
        return size, first

    def extract_then(
        then: lang.ThenNode,
    ) -> tuple[model.Field, expr.Expr, expr.Expr, expr.Expr, Location]:
        target = (
            model.FINAL
            if then.f_target.text == "null"
            else model.Field(create_id(error, then.f_target, filename))
        )
        condition = (
            create_bool_expression(error, then.f_condition, filename)
            if then.f_condition
            else expr.TRUE
        )
        size, first = extract_aspect(then.f_aspects)
        return target, condition, size, first, node_location(then, filename)

    field_identifiers = {
        # Do not store errors in `error` object as this would result in duplicate error messages
        # from the invocation of create_message_types
        i: create_id(RecordFluxError(), field.f_identifier, filename)
        for i, field in enumerate(fields.f_fields)
    }

    if fields.f_initial_field:
        structure = [
            model.Link(model.INITIAL, *extract_then(then))
            for then in fields.f_initial_field.f_thens
        ]
    else:
        structure = [
            model.Link(
                model.INITIAL,
                model.Field(field_identifiers[0]),
                location=field_identifiers[0].location,
            ),
        ]

    for i, field in enumerate(fields.f_fields):
        source_node = model.Field(field_identifiers[i]) if field.f_identifier else model.INITIAL
        if field.f_identifier.text.lower() == "message":
            error.extend(
                [
                    (
                        'reserved word "Message" used as identifier',
                        Subsystem.PARSER,
                        Severity.ERROR,
                        field_identifiers[i].location,
                    ),
                ],
            )

        if len(field.f_thens) == 0:
            target_id = field_identifiers[i + 1] if i + 1 < len(fields.f_fields) else None
            target_node = model.Field(target_id) if target_id else model.FINAL
            structure.append(
                model.Link(
                    source_node,
                    target_node,
                    location=node_location(field, filename, end_location=True),
                ),
            )

        for then in field.f_thens:
            if then.f_target.text.lower() != "null" and not any(
                then.f_target.text == c.f_identifier.text for c in fields.f_fields
            ):
                error.extend(
                    [
                        (
                            f'undefined field "{then.f_target.text}"',
                            Subsystem.PARSER,
                            Severity.ERROR,
                            node_location(then.f_target, filename) if then.f_target else None,
                        ),
                    ],
                )
                continue
            structure.append(model.Link(source_node, *extract_then(then)))

        if error.errors:
            continue

        merge_field_aspects(
            error,
            field_identifiers[i],
            structure,
            *extract_aspect(field.f_aspects),
        )
        if field.f_condition:
            error.extend(
                [
                    (
                        "short form condition is not supported anymore",
                        Subsystem.PARSER,
                        Severity.ERROR,
                        node_location(field.f_condition, filename),
                    ),
                    (
                        "add condition to all outgoing links of "
                        f'field "{field_identifiers[i]}" instead',
                        Subsystem.PARSER,
                        Severity.INFO,
                        node_location(field.f_condition, filename),
                    ),
                ],
            )

    return structure


def merge_field_aspects(
    error: RecordFluxError,
    field_identifier: ID,
    structure: Sequence[model.Link],
    size: expr.Expr,
    first: expr.Expr,
) -> None:
    if first != expr.UNDEFINED or size != expr.UNDEFINED:
        for l in (l for l in structure if l.target.identifier == field_identifier):
            if first != expr.UNDEFINED:
                if l.first == expr.UNDEFINED:
                    l.first = first
                else:
                    error.extend(
                        [
                            (
                                f'first aspect of field "{field_identifier}"'
                                " conflicts with previous"
                                " specification",
                                Subsystem.MODEL,
                                Severity.ERROR,
                                first.location,
                            ),
                            (
                                "previous specification of first",
                                Subsystem.MODEL,
                                Severity.INFO,
                                l.first.location,
                            ),
                        ],
                    )

            if size != expr.UNDEFINED:
                if l.size == expr.UNDEFINED:
                    l.size = size
                else:
                    error.extend(
                        [
                            (
                                f'size aspect of field "{field_identifier}" conflicts with'
                                " previous specification",
                                Subsystem.MODEL,
                                Severity.ERROR,
                                size.location,
                            ),
                            (
                                "previous specification of size",
                                Subsystem.MODEL,
                                Severity.INFO,
                                l.size.location,
                            ),
                        ],
                    )


def check_duplicate_aspect(
    error: RecordFluxError,
    name: str,
    locations: Sequence[Location],
) -> None:
    if len(locations) > 1:
        error.extend(
            [
                (
                    f'duplicate aspect "{name}"',
                    Subsystem.PARSER,
                    Severity.ERROR,
                    locations[1],
                ),
                *[
                    (
                        "previous location",
                        Subsystem.PARSER,
                        Severity.INFO,
                        l,
                    )
                    for l in locations[:-1]
                ],
            ],
        )


def parse_aspects(  # noqa: PLR0912
    error: RecordFluxError,
    aspects: lang.MessageAspectList,
    filename: Path,
) -> tuple[
    Mapping[ID, Sequence[expr.Expr]],
    Union[model.ByteOrder, dict[model.Field, model.ByteOrder]],
]:
    checksum_result = {}
    byte_order_result: Union[model.ByteOrder, dict[model.Field, model.ByteOrder]] = {}

    grouped = defaultdict(list)
    for aspect in aspects:
        if isinstance(aspect, lang.ByteOrderAspect):
            name = "Byte_Order"
        elif isinstance(aspect, lang.ChecksumAspect):
            name = "Checksum"
        else:
            raise NotImplementedError(f"Message aspect {type(aspect)} unsupported")
        grouped[name].append((aspect, node_location(aspect, filename)))

    for name, locations in grouped.items():
        check_duplicate_aspect(error, name, [l for _, l in locations])
        aspect, _ = locations[0]

        if isinstance(aspect, lang.ChecksumAspect):
            for assoc in aspect.f_associations:
                exprs = []
                for value in assoc.f_covered_fields:
                    if isinstance(value, lang.ChecksumVal):
                        exprs.append(create_math_expression(error, value.f_data, filename))
                    elif isinstance(value, lang.ChecksumValueRange):
                        exprs.append(
                            expr.ValueRange(
                                create_math_expression(error, value.f_first, filename),
                                create_math_expression(error, value.f_last, filename),
                            ),
                        )
                    else:
                        raise NotImplementedError(f"Invalid checksum association {value.kind_name}")
                checksum_result[create_id(error, assoc.f_identifier, filename)] = exprs
        if isinstance(aspect, lang.ByteOrderAspect):
            if isinstance(aspect.f_byte_order, lang.ByteOrderTypeLoworderfirst):
                byte_order_result = model.ByteOrder.LOW_ORDER_FIRST
            else:
                byte_order_result = model.ByteOrder.HIGH_ORDER_FIRST

    return checksum_result, byte_order_result


def create_derived_message(
    error: RecordFluxError,
    identifier: ID,
    _parameters: lang.Parameters,
    derivation: lang.TypeDef,
    filename: Path,
) -> Optional[model.UncheckedDerivedMessage]:
    assert isinstance(derivation, lang.TypeDerivationDef)
    base_id = create_id(error, derivation.f_base, filename)
    base_name = model.internal_type_identifier(base_id, identifier.parent)
    return model.UncheckedDerivedMessage(
        identifier,
        base_name,
        location=type_location(identifier, derivation),
    )


def create_enumeration(
    error: RecordFluxError,
    identifier: ID,
    _parameters: lang.Parameters,
    enumeration: lang.TypeDef,
    filename: Path,
) -> Optional[model.UncheckedEnumeration]:
    assert isinstance(enumeration, lang.EnumerationTypeDef)
    literals: list[tuple[ID, expr.Number]] = []

    def create_aspects(aspects: lang.AspectList) -> Optional[tuple[expr.Expr, bool]]:
        always_valid = False
        size = None

        grouped = defaultdict(list)
        for aspect in aspects:
            grouped[aspect.f_identifier.text].append((aspect, node_location(aspect, filename)))

        for name, locations in grouped.items():
            check_duplicate_aspect(error, name, [l for _, l in locations])
            aspect, _ = locations[0]

            if aspect.f_identifier.text == "Size":
                size = create_math_expression(error, aspect.f_value, filename)
            if aspect.f_identifier.text == "Always_Valid":
                if aspect.f_value:
                    av_expr = create_bool_expression(error, aspect.f_value, filename)
                    if av_expr == expr.TRUE:
                        always_valid = True
                    elif av_expr == expr.FALSE:
                        always_valid = False
                    else:
                        error.extend(
                            [
                                (
                                    f"invalid Always_Valid expression: {av_expr}",
                                    Subsystem.PARSER,
                                    Severity.ERROR,
                                    node_location(aspect.f_value, filename),
                                ),
                            ],
                        )
                else:
                    always_valid = True
        if not size:
            error.extend(
                [
                    (
                        f'no size set for "{identifier}"',
                        Subsystem.PARSER,
                        Severity.ERROR,
                        identifier.location,
                    ),
                ],
            )
            return None

        assert size
        return size, always_valid

    if isinstance(enumeration.f_elements, lang.NamedEnumerationDef):
        for e in enumeration.f_elements.f_elements:
            element_identifier = create_id(error, e.f_identifier, filename)
            value = create_math_expression(error, e.f_literal, filename)
            assert isinstance(value, expr.Number)
            literals.append((element_identifier, value))
    elif isinstance(enumeration.f_elements, lang.PositionalEnumerationDef):
        literals = [
            (create_id(error, e, filename), expr.Number(i))
            for i, e in enumerate(enumeration.f_elements.f_elements)
        ]
    else:
        raise NotImplementedError(
            f"Enumeration kind {enumeration.f_elements.kind_name} unsupported",
        )

    aspects = create_aspects(enumeration.f_aspects)
    if aspects is None:
        return None

    size, always_valid = aspects

    return model.UncheckedEnumeration(
        identifier,
        literals,
        size,
        always_valid,
        location=type_location(identifier, enumeration),
    )


def create_refinement(
    error: RecordFluxError,
    refinement: lang.RefinementDecl,
    package: ID,
    filename: Path,
) -> model.UncheckedRefinement:
    pdu = model.internal_type_identifier(create_id(error, refinement.f_pdu, filename), package)
    sdu = model.internal_type_identifier(create_id(error, refinement.f_sdu, filename), package)

    if refinement.f_condition:
        condition = create_bool_expression(error, refinement.f_condition, filename)
    else:
        condition = expr.TRUE

    field = model.Field(create_id(error, refinement.f_field, filename))

    return model.UncheckedRefinement(
        package,
        pdu,
        field,
        sdu,
        condition,
        node_location(refinement, filename),
    )


def check_naming(error: RecordFluxError, package: lang.PackageNode, name: Path) -> None:
    identifier = package.f_identifier.text
    if identifier.startswith("RFLX"):
        error.extend(
            [
                (
                    f'illegal prefix "RFLX" in package identifier "{identifier}"',
                    Subsystem.PARSER,
                    Severity.ERROR,
                    node_location(package.f_identifier, name),
                ),
            ],
        )
    if identifier != package.f_end_identifier.text:
        error.extend(
            [
                (
                    f'inconsistent package identifier "{package.f_end_identifier.text}"',
                    Subsystem.PARSER,
                    Severity.ERROR,
                    node_location(package.f_end_identifier, name),
                ),
                (
                    f'previous identifier was "{identifier}"',
                    Subsystem.PARSER,
                    Severity.INFO,
                    node_location(package.f_identifier, name),
                ),
            ],
        )
    if name != STDIN:
        expected_filename = f"{identifier.lower()}.rflx"
        if name.name != expected_filename:
            error.extend(
                [
                    (
                        f'file name does not match unit name "{identifier}",'
                        f' should be "{expected_filename}"',
                        Subsystem.PARSER,
                        Severity.ERROR,
                        node_location(package.f_identifier, name),
                    ),
                ],
            )


@dataclass(frozen=True)
class ContextClause:
    name: ID
    location: Location

    @property
    def withed_file(self) -> str:
        return f"{str(self.name).lower()}.rflx"


@dataclass(frozen=True)
class SpecificationFile:
    filename: Path
    spec: lang.Specification
    context_clauses: list[ContextClause]

    @staticmethod
    def create(
        error: RecordFluxError,
        spec: lang.Specification,
        filename: Path,
    ) -> SpecificationFile:
        check_naming(error, spec.f_package_declaration, filename)

        return SpecificationFile(
            filename,
            spec,
            [
                ContextClause(
                    create_id(error, context.f_item, filename),
                    node_location(context.f_item, filename),
                )
                for context in spec.f_context_clause
            ],
        )

    @property
    def package(self) -> ID:
        return ID(self.spec.f_package_declaration.f_identifier.text)


class Parser:
    def __init__(
        self,
        cache: Optional[Cache] = None,
        workers: int = 1,
        integration_files_dir: Optional[Path] = None,
    ) -> None:
        self._cache = AlwaysVerify() if cache is None else cache
        self._workers = workers
        self._specifications: OrderedDict[ID, SpecificationFile] = OrderedDict()
        self._integration: Integration = Integration(integration_files_dir)

    def parse(self, *specfiles: Path) -> None:
        error = RecordFluxError()

        include_paths = []
        specifications = []

        for f in unique(specfiles):
            if f.parent.is_dir() and f.parent not in include_paths:
                include_paths.append(f.parent)

            spec = self._parse_file(error, f)
            if spec:
                specifications.append(spec)

        _check_for_duplicate_specifications(
            error,
            [*self._specifications.values(), *specifications],
        )

        self._specifications.update({s.package: s for s in specifications})

        self._parse_withed_files(
            error,
            [c for s in specifications for c in s.context_clauses],
            include_paths,
        )

        _check_for_dependency_cycles(error, specifications, self._specifications)

        self._specifications = _sort_specs_topologically(self._specifications)

        error.propagate()

    def parse_string(
        self,
        string: str,
        filename: Path = STDIN,
        rule: str = lang.GrammarRule.main_rule_rule,
    ) -> None:
        error = RecordFluxError()
        string = textwrap.dedent(string)
        unit = lang.AnalysisContext().get_from_buffer(str(filename), string, rule=rule)

        if not diagnostics_to_error(unit.diagnostics, error, filename):
            assert isinstance(unit.root, lang.Specification)
            spec = SpecificationFile.create(error, unit.root, filename)
            _check_for_duplicate_specifications(error, [*self._specifications.values(), spec])

            self._specifications[spec.package] = spec

            _check_for_dependency_cycles(error, [spec], self._specifications)

            self._specifications = _sort_specs_topologically(self._specifications)

            error.extend(style.check_string(string, filename))

        error.propagate()

    def create_unchecked_model(self) -> model.UncheckedModel:
        error = RecordFluxError()
        declarations: list[model.UncheckedTopLevelDeclaration] = [
            model.UNCHECKED_BOOLEAN,
            model.UNCHECKED_OPAQUE,
        ]

        for spec_node in self._specifications.values():
            self._evaluate_specification(error, declarations, spec_node.spec, spec_node.filename)

        return model.UncheckedModel(declarations, error)

    def create_model(self) -> model.Model:
        unchecked_model = self.create_unchecked_model()
        error = unchecked_model.error
        checked_model = unchecked_model.checked(
            self._cache,
            self._workers,
        )
        self._integration.validate(checked_model, error)
        error.propagate()
        return checked_model

    def get_integration(self) -> Integration:
        return self._integration

    @property
    def specifications(self) -> dict[str, lang.Specification]:
        return {
            spec_node.spec.f_package_declaration.f_identifier.text: spec_node.spec
            for spec_node in self._specifications.values()
        }

    def _parse_file(self, error: RecordFluxError, filename: Path) -> Optional[SpecificationFile]:
        log.info("Parsing %s", filename)
        unit = lang.AnalysisContext().get_from_file(str(filename))

        error.extend(style.check(filename))

        if diagnostics_to_error(unit.diagnostics, error, filename):
            return None

        assert isinstance(unit.root, lang.Specification)

        self._integration.load_integration_file(filename, error)

        return SpecificationFile.create(error, unit.root, filename)

    def _parse_withed_files(
        self,
        error: RecordFluxError,
        context_clauses: Sequence[ContextClause],
        include_paths: Sequence[Path],
    ) -> None:
        if not context_clauses:
            return

        nested_context_clauses: list[ContextClause] = []
        for context_clause in context_clauses:
            if context_clause.name in self._specifications:
                continue
            for path in include_paths:
                f = path / context_clause.withed_file
                if f.exists():
                    spec = self._parse_file(error, f)
                    if spec:
                        self._specifications[spec.package] = spec
                        nested_context_clauses.extend(
                            [
                                c
                                for c in spec.context_clauses
                                if c
                                not in [
                                    *context_clauses,
                                    *nested_context_clauses,
                                ]
                                and c.name not in self._specifications
                            ],
                        )
                    break
            else:
                error.extend(
                    [
                        (
                            f'cannot find specification "{context_clause.name}"',
                            Subsystem.PARSER,
                            Severity.ERROR,
                            context_clause.location,
                        ),
                    ],
                )

        self._parse_withed_files(error, nested_context_clauses, include_paths)

    def _evaluate_specification(
        self,
        error: RecordFluxError,
        declarations: list[model.UncheckedTopLevelDeclaration],
        spec: lang.Specification,
        filename: Path,
    ) -> None:
        handlers: Mapping[
            str,
            Callable[
                [
                    RecordFluxError,
                    ID,
                    lang.Parameters,
                    lang.TypeDef,
                    Path,
                ],
                Optional[model.UncheckedType],
            ],
        ] = {
            "SequenceTypeDef": create_sequence,
            "ModularTypeDef": create_modular,
            "RangeTypeDef": create_range,
            "MessageTypeDef": create_message,
            "NullMessageTypeDef": create_null_message,
            "TypeDerivationDef": create_derived_message,
            "EnumerationTypeDef": create_enumeration,
        }
        log.info("Processing %s", spec.f_package_declaration.f_identifier.text)
        package_id = create_id(error, spec.f_package_declaration.f_identifier, filename)

        for t in spec.f_package_declaration.f_declarations:
            if isinstance(t, lang.TypeDecl):
                type_id = create_id(error, t.f_identifier, filename)
                identifier = ID(package_id * type_id, location=type_id.location)
                if t.f_definition.kind_name != "MessageTypeDef" and t.f_parameters:
                    error.extend(
                        [
                            (
                                "only message types can be parameterized",
                                Subsystem.PARSER,
                                Severity.ERROR,
                                node_location(t.f_parameters.f_parameters, filename),
                            ),
                        ],
                    )
                validate_handler(
                    error,
                    "definition",
                    t.f_definition,
                    list(handlers.keys()),
                    filename,
                )
                new_type = handlers[t.f_definition.kind_name](
                    error,
                    identifier,
                    t.f_parameters,
                    t.f_definition,
                    filename,
                )
                if new_type is not None:
                    declarations.append(new_type)
            elif isinstance(t, lang.RefinementDecl):
                declarations.append(create_refinement(error, t, package_id, filename))
            elif isinstance(t, lang.SessionDecl):
                declarations.append(create_session(error, t, package_id, filename))
            else:
                raise NotImplementedError(f"Declaration kind {t.kind_name} unsupported")


def _check_for_duplicate_specifications(
    error: RecordFluxError,
    specifications: Sequence[SpecificationFile],
) -> None:
    for i, n1 in enumerate(specifications, start=1):
        for n2 in specifications[i:]:
            if n1.package == n2.package:
                error.extend(
                    [
                        (
                            "duplicate specification",
                            Subsystem.PARSER,
                            Severity.ERROR,
                            node_location(n2.spec.f_package_declaration.f_identifier, n2.filename),
                        ),
                        (
                            "previous specification",
                            Subsystem.PARSER,
                            Severity.INFO,
                            node_location(n1.spec.f_package_declaration.f_identifier, n1.filename),
                        ),
                    ],
                )


def _check_for_dependency_cycles(
    error: RecordFluxError,
    given_specs: list[SpecificationFile],
    specifications: Mapping[ID, SpecificationFile],
) -> bool:
    result = False

    for spec in given_specs:
        for p, c in [(c.name, c) for c in spec.context_clauses]:
            try:
                _check_for_dependency_cycle(p, c, [], specifications)
            except RecordFluxError as e:  # noqa: PERF203
                result = True
                error.extend(e)

    return result


def _check_for_dependency_cycle(
    package: ID,
    context: ContextClause,
    visited: list[tuple[ID, ContextClause]],
    specifications: Mapping[ID, SpecificationFile],
) -> None:
    if package not in specifications:
        return  # ignore missing specifications

    visited_packages = [s for s, _ in visited]
    if package in visited_packages:
        idx = visited_packages.index(package)
        assert isinstance(context, ContextClause)
        cycle: list[ContextClause] = []
        for _, c in visited[idx:]:
            assert isinstance(c, ContextClause)
            cycle.append(c)
        if context not in cycle:
            cycle.append(context)
        error = RecordFluxError()
        error.extend(
            [
                (
                    f'dependency cycle when including "{cycle[0].name}"',
                    Subsystem.PARSER,
                    Severity.ERROR,
                    cycle[0].location,
                ),
                *[
                    (
                        f'when including "{c.name}"',
                        Subsystem.PARSER,
                        Severity.INFO,
                        c.location,
                    )
                    for c in cycle[1:]
                ],
            ],
        )
        raise error

    for p, c in [(c.name, c) for c in specifications[package].context_clauses]:
        _check_for_dependency_cycle(p, c, [*visited, (package, context)], specifications)


def _sort_specs_topologically(
    specifications: Mapping[ID, SpecificationFile],
) -> OrderedDict[ID, SpecificationFile]:
    """
    (Reverse) Topologically sort specifications using Kahn's algorithm.

    Specifications which are part of a cycle will be removed.
    """

    result: list[ID] = []
    incoming: dict[ID, set[ID]] = {f: set() for f in specifications}
    for package, spec_node in specifications.items():
        for c in spec_node.context_clauses:
            if c.name in incoming:
                incoming[c.name].add(package)

    specs = [f for f, i in incoming.items() if len(i) == 0]
    visited = set(specs)

    while specs:
        s = specs.pop(0)
        result.insert(0, s)
        for c in specifications[s].context_clauses:
            visited.add(c.name)
            if c.name in incoming and incoming[c.name] <= visited:
                specs.append(c.name)

    return OrderedDict((f, specifications[f]) for f in result)



from __future__ import annotations

import typing as ty
from abc import abstractmethod
from collections import abc
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import rflx.typing_ as rty
from rflx import const, expression as expr
from rflx.common import indent_next, verbose_repr
from rflx.error import Location, RecordFluxError, Severity, Subsystem, fail
from rflx.identifier import ID, StrID

from . import message
from .top_level_declaration import TopLevelDeclaration, UncheckedTopLevelDeclaration


class Type(TopLevelDeclaration):
    @property
    def type_(self) -> rty.Type:
        return rty.Undefined()

    @property
    def direct_dependencies(self) -> list[Type]:
        """
        Return a list consisting of the type and all the types on which the type directly depends.

        The dependencies are not determined recursively, i.e. the dependencies
        of dependencies are not considered.
        """
        return [self]

    @property
    def dependencies(self) -> list[Type]:
        """
        Return a list consisting of the type and all types on which the type depends.

        The dependencies are determined recursively in *postorder*.
        """
        return [self]

    @property
    def qualified_identifier(self) -> ID:
        """
        Return the qualified identifier of this type.

        The qualified identifier of a type is defined as its complete package
        path followed by the type name, or just the type name if the type is
        builtin or internal.
        """
        identifier = self.identifier
        return (
            ID(self.name, location=identifier.location)
            if is_builtin_type(identifier) or is_internal_type(identifier)
            else identifier
        )


class Scalar(Type):
    def __init__(
        self,
        identifier: StrID,
        size: expr.Expr,
        location: Optional[Location] = None,
    ) -> None:
        super().__init__(identifier, location)

        self.error.propagate()

        self._size = size

    @property
    def size(self) -> expr.Number:
        size_num = self._size.simplified()
        assert isinstance(size_num, expr.Number)
        return size_num

    @property
    def size_expr(self) -> expr.Expr:
        return self._size

    @property
    @abstractmethod
    def value_count(self) -> expr.Number:
        raise NotImplementedError

    @abstractmethod
    def constraints(
        self,
        name: str,
        proof: bool = False,
        same_package: bool = True,
    ) -> abc.Sequence[expr.Expr]:
        raise NotImplementedError


class Integer(Scalar):
    def __init__(
        self,
        identifier: StrID,
        first: expr.Expr,
        last: expr.Expr,
        size: expr.Expr,
        location: Optional[Location] = None,
    ) -> None:
        super().__init__(identifier, size, location)

        first_num = first.simplified()
        last_num = last.simplified()
        size_num = size.simplified()

        if not isinstance(first_num, expr.Number):
            self.error.extend(
                [
                    (
                        f'first of "{self.name}" contains variable',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        self.location,
                    ),
                ],
            )

        if not isinstance(last_num, expr.Number):
            self.error.fail(
                f'last of "{self.name}" contains variable',
                Subsystem.MODEL,
                Severity.ERROR,
                self.location,
            )

        if int(last_num) >= 2**const.MAX_SCALAR_SIZE:
            self.error.extend(
                [
                    (
                        f'last of "{self.name}" exceeds limit (2**{const.MAX_SCALAR_SIZE} - 1)',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        self.location,
                    ),
                ],
            )

        if not isinstance(size_num, expr.Number):
            self.error.extend(
                [
                    (
                        f'size of "{self.name}" contains variable',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        self.location,
                    ),
                ],
            )

        self.error.propagate()

        assert isinstance(first_num, expr.Number)
        assert isinstance(last_num, expr.Number)
        assert isinstance(size_num, expr.Number)

        if first_num < expr.Number(0):
            self.error.extend(
                [
                    (
                        f'first of "{self.name}" negative',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        self.location,
                    ),
                ],
            )
        if first_num > last_num:
            self.error.extend(
                [
                    (
                        f'range of "{self.name}" negative',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        self.location,
                    ),
                ],
            )

        if int(last_num).bit_length() > int(size_num):
            self.error.extend(
                [
                    (
                        f'size of "{self.name}" too small',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        self.location,
                    ),
                ],
            )

        # Eng/RecordFlux/RecordFlux#1077
        # size of integers is limited to 63bits

        if int(size_num) > const.MAX_SCALAR_SIZE:
            self.error.extend(
                [
                    (
                        f'size of "{self.name}" exceeds limit (2**{const.MAX_SCALAR_SIZE})',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        self.location,
                    ),
                ],
            )

        self.error.propagate()

        self._first_expr = first
        self._first = first_num
        self._last_expr = last
        self._last = last_num

    def __repr__(self) -> str:
        return verbose_repr(self, ["identifier", "first_expr", "last_expr", "size_expr"])

    def __str__(self) -> str:
        return (
            f"type {self.name} is range {self.first_expr} .. {self.last_expr}"
            f" with Size => {self.size_expr}"
        )

    @property
    def type_(self) -> rty.Type:
        return rty.Integer(
            self.full_name,
            rty.Bounds(self.first.value, self.last.value),
            location=self.location,
        )

    @property
    def value_count(self) -> expr.Number:
        return self.last - self.first + expr.Number(1)

    @property
    def first(self) -> expr.Number:
        return self._first

    @property
    def first_expr(self) -> expr.Expr:
        return self._first_expr

    @property
    def last(self) -> expr.Number:
        return self._last

    @property
    def last_expr(self) -> expr.Expr:
        return self._last_expr

    def constraints(
        self,
        name: str,
        proof: bool = False,  # noqa: ARG002
        same_package: bool = True,  # noqa: ARG002
    ) -> abc.Sequence[expr.Expr]:
        return [
            expr.GreaterEqual(
                expr.Variable(name, type_=self.type_),
                self.first,
                location=self.location,
            ),
            expr.LessEqual(
                expr.Variable(name, type_=self.type_),
                self.last,
                location=self.location,
            ),
            expr.Equal(expr.Size(name), self.size, location=self.location),
        ]


class Enumeration(Scalar):
    def __init__(  # noqa: PLR0912
        self,
        identifier: StrID,
        literals: abc.Sequence[tuple[StrID, expr.Number]],
        size: expr.Expr,
        always_valid: bool,
        location: Optional[Location] = None,
    ) -> None:
        super().__init__(identifier, size, location)

        for i1, e1 in enumerate(literals):
            for i2, e2 in enumerate(literals):
                if i2 < i1 and e1[0] == e2[0]:
                    self.error.extend(
                        [
                            (
                                f'duplicate literal "{e1[0]}"',
                                Subsystem.MODEL,
                                Severity.ERROR,
                                e1[0].location if isinstance(e1[0], ID) else self.location,
                            ),
                            (
                                "previous occurrence",
                                Subsystem.MODEL,
                                Severity.INFO,
                                e2[0].location if isinstance(e2[0], ID) else self.location,
                            ),
                        ],
                    )

        self.literals = {}
        for k, v in literals:
            if " " in str(k) or "." in str(k):
                self.error.extend(
                    [
                        (
                            f'invalid literal name "{k}" in "{self.name}"',
                            Subsystem.MODEL,
                            Severity.ERROR,
                            self.location,
                        ),
                    ],
                )
                continue
            self.literals[ID(k)] = v

        size_num = size.simplified()

        if not isinstance(size_num, expr.Number):
            self.error.fail(
                f'size of "{self.name}" contains variable',
                Subsystem.MODEL,
                Severity.ERROR,
                self.location,
            )

        if self.literals.values():
            min_literal_value = min(map(int, self.literals.values()))
            max_literal_value = max(map(int, self.literals.values()))
            if min_literal_value < 0 or max_literal_value > 2**const.MAX_SCALAR_SIZE - 1:
                self.error.extend(
                    [
                        (
                            f'enumeration value of "{self.name}"'
                            f" outside of permitted range (0 .. 2**{const.MAX_SCALAR_SIZE} - 1)",
                            Subsystem.MODEL,
                            Severity.ERROR,
                            self.location,
                        ),
                    ],
                )
            if max_literal_value.bit_length() > int(size_num):
                self.error.extend(
                    [
                        (
                            f'size of "{self.name}" too small',
                            Subsystem.MODEL,
                            Severity.ERROR,
                            self.location,
                        ),
                    ],
                )

        # Eng/RecordFlux/RecordFlux#1077
        # size of integers is limited to 63bits

        if int(size_num) > const.MAX_SCALAR_SIZE:
            self.error.extend(
                [
                    (
                        f'size of "{self.name}" exceeds limit (2**{const.MAX_SCALAR_SIZE})',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        self.location,
                    ),
                ],
            )
        for i1, v1 in enumerate(self.literals.values()):
            for i2, v2 in enumerate(self.literals.values()):
                if i1 < i2 and v1 == v2:
                    self.error.extend(
                        [
                            (
                                f'duplicate enumeration value "{v1}" in "{self.name}"',
                                Subsystem.MODEL,
                                Severity.ERROR,
                                v2.location,
                            ),
                            ("previous occurrence", Subsystem.MODEL, Severity.INFO, v1.location),
                        ],
                    )

        if always_valid and len(self.literals) == 2 ** int(size_num):
            self.error.extend(
                [
                    (
                        f'unnecessary always-valid aspect on "{self.name}"',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        self.location,
                    ),
                ],
            )

        self.error.propagate()

        self.always_valid = always_valid

    def __repr__(self) -> str:
        return verbose_repr(self, ["identifier", "literals", "size_expr", "always_valid"])

    def __str__(self) -> str:
        literals = ", ".join(f"{l} => {v}" for l, v in self.literals.items())
        always_valid = f", Always_Valid => {self.always_valid}" if self.always_valid else ""
        aspects = f"with Size => {self.size_expr}{always_valid}"
        result = f"type {self.name} is ({literals}) {aspects}"

        if len(result) > 100:
            literals = ",\n".join(f"{l} => {v}" for l, v in self.literals.items())
            result = f"type {self.name} is\n   ({indent_next(literals, 4)})\n{aspects}"

        return result

    @property
    def type_(self) -> rty.Type:
        return rty.Enumeration(
            self.full_name,
            list(map(ID, self.literals.keys())),
            self.always_valid,
            location=self.location,
        )

    @property
    def value_count(self) -> expr.Number:
        if self.always_valid:
            size_num = self.size.simplified()
            assert isinstance(size_num, expr.Number)
            return expr.Number(2 ** int(size_num))
        return expr.Number(len(self.literals))

    def constraints(
        self,
        name: str,
        proof: bool = False,
        same_package: bool = True,
    ) -> abc.Sequence[expr.Expr]:
        literals = dict(self.literals)

        if proof and self.always_valid:
            # Add a dummy literal to indicate that there are other valid values,
            # thus preventing a condition from being falsely detected as always true.
            for v in range(2**self.size.value):  # pragma: no branch
                if expr.Number(v) not in literals.values():
                    literals["L_" + self.identifier.name + f"_{v}"] = expr.Number(v)
                    break

        prefixed_literals = {self.package * l: v for l, v in literals.items()}
        if self.package == const.BUILTINS_PACKAGE:
            pass
        elif same_package:
            literals = {**literals, **prefixed_literals}
        else:
            literals = prefixed_literals

        result: list[expr.Expr] = [
            expr.Or(
                *[
                    expr.Equal(
                        expr.Variable(name, type_=self.type_),
                        expr.Literal(l),
                        self.location,
                    )
                    for l in literals
                ],
                location=self.location,
            ),
        ]
        result.extend(
            [
                expr.Equal(expr.Literal(l, type_=self.type_), v, self.location)
                for l, v in literals.items()
            ],
        )
        result.append(expr.Equal(expr.Size(name), self.size, self.location))
        return result


class Composite(Type):
    @property
    @abstractmethod
    def element_size(self) -> expr.Expr:
        raise NotImplementedError


class Sequence(Composite):
    def __init__(
        self,
        identifier: StrID,
        element_type: Type,
        location: Optional[Location] = None,
    ) -> None:
        super().__init__(identifier, location)
        self.element_type = element_type

        if not isinstance(element_type, Scalar) and not isinstance(element_type, message.Message):
            self.error.extend(
                [
                    (
                        f'invalid element type of sequence "{self.name}"',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        location,
                    ),
                    (
                        f'type "{element_type.name}" must be scalar or message',
                        Subsystem.MODEL,
                        Severity.INFO,
                        element_type.location,
                    ),
                ],
            )

        if isinstance(element_type, Scalar):
            element_type_size = element_type.size.simplified()
            if not isinstance(element_type_size, expr.Number) or int(element_type_size) % 8 != 0:
                self.error.extend(
                    [
                        (
                            f'unsupported element type size of sequence "{self.name}"',
                            Subsystem.MODEL,
                            Severity.ERROR,
                            location,
                        ),
                        (
                            f'type "{element_type.name}" has size {element_type_size},'
                            r" must be multiple of 8",
                            Subsystem.MODEL,
                            Severity.INFO,
                            element_type.location,
                        ),
                    ],
                )

        if isinstance(element_type, message.Message):
            error_entry = (
                f'invalid element type of sequence "{self.name}"',
                Subsystem.MODEL,
                Severity.ERROR,
                location,
            )

            if element_type.is_null:
                self.error.extend(
                    [
                        error_entry,
                        (
                            "null messages must not be used as sequence element",
                            Subsystem.MODEL,
                            Severity.INFO,
                            element_type.location,
                        ),
                    ],
                )

            if element_type.parameters:
                self.error.extend(
                    [
                        error_entry,
                        (
                            "parameterized messages must not be used as sequence element",
                            Subsystem.MODEL,
                            Severity.INFO,
                            element_type.location,
                        ),
                    ],
                )

            if any(
                bool(e.findall(lambda x: x in [expr.Size("Message"), expr.Last("Message")]))
                for l in element_type.structure
                for e in [l.condition, l.size]
            ):
                self.error.extend(
                    [
                        error_entry,
                        (
                            "messages used as sequence element must not depend"
                            ' on "Message\'Size" or "Message\'Last"',
                            Subsystem.MODEL,
                            Severity.INFO,
                            element_type.location,
                        ),
                    ],
                )

        self.error.propagate()

    def __repr__(self) -> str:
        return verbose_repr(self, ["identifier", "element_type"])

    def __str__(self) -> str:
        return f"type {self.name} is sequence of {self.element_type.qualified_identifier}"

    @property
    def type_(self) -> rty.Type:
        return rty.Sequence(self.full_name, self.element_type.type_)

    @property
    def element_size(self) -> expr.Expr:
        return expr.Size(self.element_type.name)

    @property
    def direct_dependencies(self) -> list[Type]:
        return [self.element_type, self]

    @property
    def dependencies(self) -> list[Type]:
        return [*self.element_type.dependencies, self]


class Opaque(Composite):
    def __init__(self, location: Optional[Location] = None) -> None:
        super().__init__(const.INTERNAL_PACKAGE * "Opaque", location)

    def __repr__(self) -> str:
        return verbose_repr(self, [])

    def __str__(self) -> str:
        return ""

    @property
    def type_(self) -> rty.Type:
        return rty.OPAQUE

    @property
    def element_size(self) -> expr.Expr:
        return expr.Number(8)


@dataclass
class UncheckedType(UncheckedTopLevelDeclaration):
    identifier: ID

    @abstractmethod
    def checked(
        self,
        declarations: ty.Sequence[TopLevelDeclaration],
        skip_verification: bool = False,
        workers: int = 1,
    ) -> Type:
        raise NotImplementedError


@dataclass
class UncheckedInteger(UncheckedType):
    identifier: ID
    first: expr.Expr
    last: expr.Expr
    size: expr.Expr
    location: Optional[Location]

    def checked(
        self,
        _declarations: ty.Sequence[TopLevelDeclaration],
        skip_verification: bool = False,  # noqa: ARG002
        workers: int = 1,  # noqa: ARG002
    ) -> Integer:
        return Integer(self.identifier, self.first, self.last, self.size, self.location)


@dataclass
class UncheckedEnumeration(UncheckedType):
    identifier: ID
    literals: abc.Sequence[tuple[ID, expr.Number]]
    size: expr.Expr
    always_valid: bool
    location: Optional[Location]

    def checked(
        self,
        _declarations: ty.Sequence[TopLevelDeclaration],
        skip_verification: bool = False,  # noqa: ARG002
        workers: int = 1,  # noqa: ARG002
    ) -> Enumeration:
        return Enumeration(
            self.identifier,
            self.literals,
            self.size,
            self.always_valid,
            self.location,
        )


@dataclass
class UncheckedSequence(UncheckedType):
    identifier: ID
    element_identifier: ID
    location: Optional[Location]

    def checked(
        self,
        declarations: ty.Sequence[TopLevelDeclaration],
        skip_verification: bool = False,  # noqa: ARG002
        workers: int = 1,  # noqa: ARG002
    ) -> Sequence:
        element_type = next(
            (
                t
                for t in declarations
                if isinstance(t, Type) and t.identifier == self.element_identifier
            ),
            None,
        )
        if not element_type:
            fail(
                f'undefined element type "{self.element_identifier}"',
                Subsystem.MODEL,
                Severity.ERROR,
                self.element_identifier.location,
            )
        return Sequence(self.identifier, element_type, self.location)


@dataclass
class UncheckedOpaque(UncheckedType):
    identifier: ID
    location: Optional[Location]

    def checked(
        self,
        _declarations: ty.Sequence[TopLevelDeclaration],
        skip_verification: bool = False,  # noqa: ARG002
        workers: int = 1,  # noqa: ARG002
    ) -> Opaque:
        return Opaque(location=Location((0, 0), Path(str(const.BUILTINS_PACKAGE)), (0, 0)))


UNCHECKED_OPAQUE = UncheckedOpaque(
    const.INTERNAL_PACKAGE * "Opaque",
    Location((0, 0), Path(str(const.BUILTINS_PACKAGE)), (0, 0)),
)
OPAQUE = UNCHECKED_OPAQUE.checked([])

INTERNAL_TYPES = {
    OPAQUE.identifier: OPAQUE,
}

UNCHECKED_BOOLEAN = UncheckedEnumeration(
    const.BUILTINS_PACKAGE * "Boolean",
    [
        (ID("False", Location((0, 0), Path(str(const.BUILTINS_PACKAGE)), (0, 0))), expr.Number(0)),
        (ID("True", Location((0, 0), Path(str(const.BUILTINS_PACKAGE)), (0, 0))), expr.Number(1)),
    ],
    expr.Number(1),
    always_valid=False,
    location=rty.BOOLEAN.location,
)
BOOLEAN = UNCHECKED_BOOLEAN.checked([])

BUILTIN_TYPES = {
    BOOLEAN.identifier: BOOLEAN,
}

BUILTIN_LITERALS = {l for t in BUILTIN_TYPES.values() for l in t.literals}


def is_internal_type(identifier: StrID) -> bool:
    return ID(identifier) in INTERNAL_TYPES or any(
        ID(identifier) == ID(t.name) for t in INTERNAL_TYPES.values()
    )


def is_builtin_type(identifier: StrID) -> bool:
    return ID(identifier) in BUILTIN_TYPES or any(
        ID(identifier) == ID(t.name) for t in BUILTIN_TYPES.values()
    )


def internal_type_identifier(identifier: ID, package: Optional[ID] = None) -> ID:
    """
    Return the internal identifier of a type.

    The internal identifier is defined as its complete package path
    (`__BUILTINS__` for builtin types, and `__INTERNAL__` for internal types)
    followed by the type name.
    """
    if is_builtin_type(identifier):
        return ID(const.BUILTINS_PACKAGE * identifier.name, location=identifier.location)

    if is_internal_type(identifier):
        return ID(const.INTERNAL_PACKAGE * identifier.name, location=identifier.location)

    if len(identifier.parts) == 1 and package:
        return ID(package * identifier, location=identifier.location)

    return identifier


def enum_literals(types: abc.Iterable[Type], package: ID) -> dict[ID, Enumeration]:
    literals = {}

    for t in types:
        if isinstance(t, Enumeration):
            for l in t.literals:
                if t.package in [const.BUILTINS_PACKAGE, package]:
                    literals[l] = t
                if t.package != const.BUILTINS_PACKAGE:
                    literals[t.package * l] = t

    return literals


def unqualified_enum_literals(types: abc.Iterable[Type], package: ID) -> dict[ID, Enumeration]:
    return {
        l: t
        for t in types
        if isinstance(t, Enumeration) and t.package == package
        for l in t.literals
    }


def qualified_enum_literals(types: abc.Iterable[Type]) -> dict[ID, Enumeration]:
    return {
        l if t.package == const.BUILTINS_PACKAGE else t.package * l: t
        for t in types
        if isinstance(t, Enumeration)
        for l in t.literals
    }


def qualified_type_names(types: abc.Iterable[Type]) -> dict[ID, Type]:
    return {
        t.identifier.name if t.package == const.BUILTINS_PACKAGE else t.identifier: t for t in types
    }


def check_identifier_notation(
    expressions: abc.Iterable[expr.Expr],
    identifiers: abc.Iterable[ID],
) -> RecordFluxError:
    id_map = {i: i for i in identifiers}

    def verify_identifier_notation(expression: expr.Expr, error: RecordFluxError) -> expr.Expr:
        if (
            isinstance(expression, (expr.Variable, expr.Literal, expr.TypeName, expr.Call))
            and expression.identifier in id_map
            and str(expression.identifier) != str(id_map[expression.identifier])
        ):
            declaration_location = id_map[expression.identifier].location
            error.extend(
                [
                    (
                        f'casing of "{expression.identifier}" differs from casing in the'
                        f' declaration of "{id_map[expression.identifier]}"'
                        + (f" at {declaration_location.short}" if declaration_location else ""),
                        Subsystem.MODEL,
                        Severity.ERROR,
                        expression.identifier.location,
                    ),
                    (
                        f'declaration of "{id_map[expression.identifier]}"',
                        Subsystem.MODEL,
                        Severity.INFO,
                        declaration_location,
                    ),
                ],
            )
        return expression

    error = RecordFluxError()

    for e in expressions:
        e.substituted(lambda e: verify_identifier_notation(e, error))

    return error



from pathlib import Path
from typing import Final

import importlib_resources

from rflx import ada
from rflx.common import file_name
from rflx.identifier import ID

REFINEMENT_PACKAGE = ID("Contains")

ARITHMETIC_PACKAGE = ID("RFLX_Arithmetic")
BUILTIN_TYPES_CONVERSIONS_PACKAGE = ID("RFLX_Builtin_Types.Conversions")
BUILTIN_TYPES_PACKAGE = ID("RFLX_Builtin_Types")
GENERIC_TYPES_PACKAGE = ID("RFLX_Generic_Types")
GENERIC_TYPES_GENERIC_OPERATORS_PACKAGE = ID("RFLX_Generic_Types.Generic_Operators")
GENERIC_TYPES_GENERIC_OPERATIONS_PACKAGE = ID("RFLX_Generic_Types.Generic_Operations")
MESSAGE_SEQUENCE_PACKAGE = ID("RFLX_Message_Sequence")
SCALAR_SEQUENCE_PACKAGE = ID("RFLX_Scalar_Sequence")
TYPES_PACKAGE = ID("RFLX_Types")
TYPES_OPERATORS_PACKAGE = ID("RFLX_Types.Operators")
TYPES_OPERATIONS_PACKAGE = ID("RFLX_Types.Operations")

LIBRARY_FILES = [
    file_name(str(p)) + ".ads"
    for p in [
        ARITHMETIC_PACKAGE,
        BUILTIN_TYPES_CONVERSIONS_PACKAGE,
        BUILTIN_TYPES_PACKAGE,
        GENERIC_TYPES_PACKAGE,
        GENERIC_TYPES_GENERIC_OPERATORS_PACKAGE,
        GENERIC_TYPES_GENERIC_OPERATIONS_PACKAGE,
        MESSAGE_SEQUENCE_PACKAGE,
        SCALAR_SEQUENCE_PACKAGE,
        TYPES_PACKAGE,
        TYPES_OPERATORS_PACKAGE,
        TYPES_OPERATIONS_PACKAGE,
    ]
] + [
    file_name(str(p)) + ".adb"
    for p in [
        ARITHMETIC_PACKAGE,
        GENERIC_TYPES_GENERIC_OPERATIONS_PACKAGE,
        MESSAGE_SEQUENCE_PACKAGE,
        SCALAR_SEQUENCE_PACKAGE,
    ]
]

# TODO(eng/recordflux/RecordFlux#1359): Replace importlib_resources by importlib.resources
TEMPLATE_DIR: Final[Path] = importlib_resources.files("rflx") / "templates"

TYPES = TYPES_PACKAGE
TYPES_BYTE = TYPES * "Byte"
TYPES_BYTES = TYPES * "Bytes"
TYPES_BYTES_PTR = TYPES * "Bytes_Ptr"
TYPES_INDEX = TYPES * "Index"
TYPES_LENGTH = TYPES * "Length"
TYPES_BIT_INDEX = TYPES * "Bit_Index"
TYPES_BIT_LENGTH = TYPES * "Bit_Length"
TYPES_TO_INDEX = TYPES * "To_Index"
TYPES_TO_LENGTH = TYPES * "To_Length"
TYPES_TO_BIT_LENGTH = TYPES * "To_Bit_Length"
TYPES_TO_FIRST_BIT_INDEX = TYPES * "To_First_Bit_Index"
TYPES_TO_LAST_BIT_INDEX = TYPES * "To_Last_Bit_Index"
TYPES_OFFSET = TYPES * "Offset"
TYPES_U64 = TYPES * "U64"
TYPES_BASE_INT = TYPES * "Base_Integer"
TYPES_BYTE_ORDER = TYPES * "Byte_Order"
TYPES_HIGH_ORDER_FIRST = TYPES * "High_Order_First"
TYPES_LOW_ORDER_FIRST = TYPES * "Low_Order_First"
TYPES_OPERATIONS = TYPES_OPERATIONS_PACKAGE

UNREACHABLE = ada.Call(TYPES * "Unreachable")

CONFIGURATION_PRAGMAS = [
    ada.Pragma("Ada_2012"),
    ada.Pragma("Style_Checks", [ada.String("N3aAbCdefhiIklnOprStux")]),
    # Eng/RecordFlux/RecordFlux#508
    ada.Pragma("Warnings", [ada.Variable("Off"), ada.String("redundant conversion")]),
]



# type: ignore # noqa: PGH003

"""RecordFlux support for GNAT Studio."""

import json
import re
from pathlib import Path

import GPS
import highlighter.common as hl
from gi.repository import Gdk, Gtk

try:
    from gs_utils import hook
except ImportError:
    from gps_utils import hook

XML = r"""<?xml version="1.0"?>
<GNAT_Studio>

   <project_attribute
      name="Output_Dir"
      package="RecordFlux"
      editor_page="RecordFlux"
      editor_section="Directories"
      description="Output directory for files generated by RecordFlux">
   </project_attribute>

   <Language>
      <Name>RecordFlux</Name>
      <Spec_Suffix>.rflx</Spec_Suffix>

      <!-- Dummy body suffix as file templates don't work when its missing -->
      <Body_Suffix>.workaround_rflx</Body_Suffix>

      <Categories>
         <Category>
            <Name>package</Name>
            <Pattern>^[ \\t]*package[ \\t]+((\\w|\\.)+)</Pattern>
            <Index>1</Index>
         </Category>
         <Category>
            <Name>type</Name>
               <Pattern>^[ \\t]*type[ \\t]+(\\w+)</Pattern>
               <Index>1</Index>
         </Category>
      </Categories>

      <Context>
         <New_Line_Comment_Start>--</New_Line_Comment_Start>
         <String_Delimiter>&quot;</String_Delimiter>
         <Can_Indent>True</Can_Indent>
         <Syntax_Highlighting>True</Syntax_Highlighting>
         <Case_Sensitive>False</Case_Sensitive>
      </Context>
   </Language>

   <!-- Filter -->
   <filter name="RecordFlux" language="RecordFlux"/>

   <!-- Actions -->
   <action name="rflx_check">
      <filter_and>
         <filter id="RecordFlux"/>
         <filter id="Source editor"/>
      </filter_and>
      <shell lang="python" show-command="false">recordflux.check(&quot;%F&quot;)</shell>
      <external>%1</external>
      <on-failure>
         <shell lang="python" show-command="false">recordflux.parse_output(&quot;&quot;&quot;%1&quot;&quot;&quot;)</shell>
      </on-failure>
   </action>

   <action name="rflx_check_all">
      <filter id="File"/>
      <shell lang="python" show-command="false">recordflux.check_all()</shell>
      <external>%1</external>
      <on-failure>
         <shell lang="python" show-command="false">recordflux.parse_output(&quot;&quot;&quot;%1&quot;&quot;&quot;)</shell>
      </on-failure>
   </action>

   <action name="rflx_generate">
      <filter_and>
         <filter id="RecordFlux"/>
         <filter id="Source editor"/>
      </filter_and>
      <shell lang="python" show-command="false">recordflux.generate(&quot;%F&quot;)</shell>
      <external>%1</external>
      <on-failure>
         <shell lang="python" show-command="false">recordflux.parse_output(&quot;&quot;&quot;%1&quot;&quot;&quot;)</shell>
      </on-failure>
   </action>

   <action name="rflx_generate_all">
      <filter id="File"/>
      <shell lang="python" show-command="false">recordflux.generate_all()</shell>
      <external>%1</external>
      <on-failure>
         <shell lang="python" show-command="false">recordflux.parse_output(&quot;&quot;&quot;%1&quot;&quot;&quot;)</shell>
      </on-failure>
   </action>

   <action name="rflx_display_graph">
      <filter id="RecordFlux"/>
      <shell lang="python" show-command="false">recordflux.graph(&quot;%F&quot;, True)</shell>
      <external>%1</external>
      <on-failure>
         <shell lang="python" show-command="false">recordflux.parse_output(&quot;&quot;&quot;%1&quot;&quot;&quot;)</shell>
      </on-failure>
      <shell lang="python" show-command="false">recordflux.display_message_graph(&quot;%F&quot;)</shell>
   </action>

   <!-- Aliases -->
   <alias name="rflx_package">
      <param name="name" description="The name of the RecordFlux package"/>
      <text>package %(name) is
   %_
end %(name);</text>
   </alias>

   <!-- Submenu -->
   <submenu>
      <title>RecordFlux</title>

      <menu after="SPARK" action="rflx_check">
         <title>Check</title>
      </menu>

      <menu action="rflx_check_all">
         <title>Check All</title>
      </menu>

      <menu action="rflx_generate">
         <title>Generate</title>
      </menu>

      <menu action="rflx_generate_all">
         <title>Generate All</title>
      </menu>
   </submenu>

    <!-- Context menu -->
    <contextual action="rflx_display_graph">
        <Title>RecordFlux/Display message graph</Title>
    </contextual>

</GNAT_Studio>
"""  # noqa: E501

GPS.parse_xml(XML)

# Highlighting
recordflux_keywords = [
    "and",
    "sequence of",
    "end",
    "for",
    "if",
    "is",
    "message",
    "mod",
    "new",
    "null",
    "or",
    "package",
    "range",
    "then",
    "type",
    "use",
    "with",
]

recordflux_literals = [
    "False",
    "True",
]

tag_aspect = hl.existing_style("Src-Editor-Aspects-Variant", "aspects")

hl_type = hl.simple(r"\b[^;\s]+", tag=hl.tag_type)
type_region = hl.region(r":", r"\b", highlighter=(hl_type,), tag=hl.tag_default)
string_literal = hl.region(r'"', r'"', matchall=False, tag=hl.tag_string)

hl.register_highlighter(
    language="recordflux",
    spec=(
        hl.simple(r"--[^\n]*", tag=hl.tag_comment),
        type_region,
        hl.simple(r"\b'First", tag=tag_aspect),
        hl.simple(r"\b'Last", tag=tag_aspect),
        hl.simple(r"\b'Length", tag=tag_aspect),
        hl.simple(r"\b'Size", tag=tag_aspect),
        hl.simple(r"\bFirst\s+=>", tag=tag_aspect),
        hl.simple(r"\bLast\s+=>", tag=tag_aspect),
        hl.simple(r"\bLength\s+=>", tag=tag_aspect),
        hl.simple(r"\bSize\s+=>", tag=tag_aspect),
        hl.words(recordflux_keywords, tag=hl.tag_keyword),
        hl.words(recordflux_literals, tag=hl.tag_keyword),
        hl.simple(r"16#[_A-Fa-f0-9]+#", tag=hl.tag_number),
        hl.simple(r"\b[_0-9]+\b", tag=hl.tag_number),
        hl.words(("Always_Valid"), tag=tag_aspect),
        string_literal,
    ),
)


@hook("gps_started")
def __on_gps_started():
    GPS.FileTemplate.register(
        alias_name="rflx_package",
        label="RecordFlux specification",
        unit_param="name",
        language="recordflux",
        is_impl=False,
    )


message_re = re.compile(
    r"^"
    r"(?P<filename>[^:]+):"
    r"(?P<line>\d+):"
    r"(?P<column>\d+): "
    r"(?P<subsystem>\w+): "
    r"(?P<severity>info|warning|error): "
    r"(?P<message>.*)"
    r"$",
)

generic_message_re = re.compile(
    r"^(?P<subsystem>\w+): (?P<severity>info|warning|error): (?P<message>.*)$",
)


def to_importance(severity):
    if severity == "error":
        return GPS.Message.Importance.HIGH
    if severity == "warning":
        return GPS.Message.Importance.MEDIUM
    if severity == "info":
        return GPS.Message.Importance.INFORMATIONAL


def parse_output(output):
    for m in GPS.Message.list():
        m.remove()
    message = None
    for l in output.splitlines():
        result = message_re.match(l)
        if result:
            data = result.groupdict()
            if data["severity"] != "info":
                message = GPS.Message(
                    category="RecordFlux",
                    file=GPS.File(data["filename"]),
                    line=int(data["line"]),
                    column=int(data["column"]),
                    text="{subsystem}: {severity}: {message}".format(**data),
                    importance=to_importance(data["severity"]),
                )
            elif message:
                data["relative"] = Path.name(data["filename"])
                message.create_nested_message(
                    file=GPS.File(data["filename"]),
                    line=int(data["line"]),
                    column=int(data["column"]),
                    text="{message} ({relative}:{line}:{column})".format(**data),
                )
            continue

        result = generic_message_re.match(l)
        if result:
            data = result.groupdict()
            message = GPS.Message(
                category="RecordFlux",
                file=GPS.File(data["subsystem"]),
                line=1,
                column=1,
                text="{subsystem}: {severity}: {message}".format(**data),
                importance=to_importance(data["severity"]),
            )


def get_source_files():
    files = (
        GPS.current_context().project().sources()
        if GPS.current_context().project()
        else GPS.current_context().files()
    )

    if files:
        return [s.name() for s in files if s.language() == "recordflux"]

    raise GPS.Exception("No files found")


def run(files, mode, skip_verification=False, options=None):
    assert mode in ("check", "generate", "graph")
    options = options or []

    GPS.MDI.save_all(force=True)
    GPS.Locations.remove_category("RecordFlux")

    return "rflx {skip_verification}{mode} {options} {files}".format(
        skip_verification="--unsafe --no-verification " if skip_verification else "",
        mode=mode,
        files=" ".join(files),
        options=" ".join(options),
    )


def check_all(filenames=None):
    filenames = filenames or get_source_files()
    return run(filenames, mode="check")


def check(filename):
    return check_all([filename])


def output_dir():
    output_dir = GPS.Project.root().get_attribute_as_string("Output_Dir", "RecordFlux")
    assert output_dir, "Output directory not configured"
    create_missing = GPS.Project.root().get_attribute_as_string("Create_Missing_Dirs")
    if create_missing and create_missing.lower() == "true":
        output_path = Path(output_dir)
        if output_path.exists():
            assert output_path.is_dir()
        else:
            output_path.mkdir(parents=True)
    return output_dir


def generate_all(filenames=None):
    filenames = filenames or get_source_files()
    return run(
        filenames,
        mode="generate",
        options=["-d", output_dir(), "--ignore-unsupported-checksum"],
    )


def generate(filename):
    return generate_all([filename])


def graph(filename, unverified=False):
    options = ["-d", output_dir()]
    return run([filename], mode="graph", skip_verification=unverified, options=options)


def get_message_name(locations, name, line, column):
    for message, pos in locations[name].items():
        if line > pos["start"]["line"] and line < pos["end"]["line"]:
            return message
        if line == pos["start"]["line"] and column >= pos["start"]["column"]:
            return message
        if line == pos["end"]["line"] and column <= pos["end"]["column"]:
            return message
    return None


def display_message_graph(filename):
    with (Path(output_dir()) / "locations.json").open() as f:
        locations = json.load(f)
    loc = GPS.EditorBuffer.get().current_view().cursor()
    column = loc.column()
    line = loc.line()
    name = GPS.File(filename).name()
    message_name = get_message_name(locations, name, line, column)
    if not message_name:
        GPS.MDI.dialog(
            f"No message found at {name}:{line}:{column}",
            "Error opening graph",
        )
        return

    scrolled_window = Gtk.ScrolledWindow()
    scrolled_window.set_border_width(0)
    scrolled_window.set_policy(Gtk.PolicyType.ALWAYS, Gtk.PolicyType.ALWAYS)
    scrolled_window.override_background_color(Gtk.StateType.NORMAL, Gdk.RGBA())
    graph = Gtk.Image()
    graph.set_from_file(output_dir() + "/" + message_name + ".svg")
    scrolled_window.add_with_viewport(graph)
    GPS.MDI.add(
        scrolled_window,
        "Message graph",
        message_name,
        group=GPS.MDI.GROUP_GRAPHS,
        position=GPS.MDI.POSITION_AUTOMATIC,
    )



from .cache import (
    AlwaysVerify as AlwaysVerify,
    Cache as Cache,
    Digest as Digest,
    NeverVerify as NeverVerify,
)
from .message import (
    FINAL as FINAL,
    INITIAL as INITIAL,
    ByteOrder as ByteOrder,
    DerivedMessage as DerivedMessage,
    Field as Field,
    Link as Link,
    Message as Message,
    Refinement as Refinement,
    UncheckedDerivedMessage as UncheckedDerivedMessage,
    UncheckedMessage as UncheckedMessage,
    UncheckedRefinement as UncheckedRefinement,
)
from .model import Model as Model, UncheckedModel as UncheckedModel
from .session import (
    FINAL_STATE as FINAL_STATE,
    Session as Session,
    State as State,
    Transition as Transition,
    UncheckedSession as UncheckedSession,
)
from .top_level_declaration import (
    TopLevelDeclaration as TopLevelDeclaration,
    UncheckedTopLevelDeclaration as UncheckedTopLevelDeclaration,
)
from .type_ import (
    BOOLEAN as BOOLEAN,
    BUILTIN_LITERALS as BUILTIN_LITERALS,
    BUILTIN_TYPES as BUILTIN_TYPES,
    INTERNAL_TYPES as INTERNAL_TYPES,
    OPAQUE as OPAQUE,
    UNCHECKED_BOOLEAN as UNCHECKED_BOOLEAN,
    UNCHECKED_OPAQUE as UNCHECKED_OPAQUE,
    Composite as Composite,
    Enumeration as Enumeration,
    Integer as Integer,
    Opaque as Opaque,
    Scalar as Scalar,
    Sequence as Sequence,
    Type as Type,
    UncheckedEnumeration as UncheckedEnumeration,
    UncheckedInteger as UncheckedInteger,
    UncheckedOpaque as UncheckedOpaque,
    UncheckedSequence as UncheckedSequence,
    UncheckedType as UncheckedType,
    internal_type_identifier as internal_type_identifier,
    is_builtin_type as is_builtin_type,
    is_internal_type as is_internal_type,
)



from __future__ import annotations

import logging
from collections import abc
from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass
from datetime import date
from functools import cached_property
from pathlib import Path
from typing import Optional

from rflx import __version__, expression as expr
from rflx.ada import (
    FALSE,
    TRUE,
    Add,
    Aggregate,
    AlwaysTerminates,
    And,
    AndThen,
    Aspect,
    Call,
    CallStatement,
    Case,
    ChoiceList,
    Component,
    Constrained,
    ContextItem,
    Declaration,
    Discriminant,
    EnumerationType,
    Equal,
    Expr,
    ExpressionFunctionDeclaration,
    First,
    FormalDeclaration,
    FunctionSpecification,
    GenericPackageInstantiation,
    GreaterEqual,
    If,
    In,
    InOutParameter,
    InstantiationUnit,
    Last,
    Less,
    Not,
    NotIn,
    Number,
    ObjectDeclaration,
    Old,
    OutParameter,
    PackageBody,
    PackageDeclaration,
    PackageUnit,
    Parameter,
    Postcondition,
    Pow,
    Pragma,
    PragmaStatement,
    Precondition,
    ProcedureSpecification,
    RangeType,
    RecordType,
    Selected,
    SizeAspect,
    Slice,
    SparkMode,
    String,
    Subprogram,
    SubprogramBody,
    SubprogramDeclaration,
    SubprogramUnitPart,
    Unit,
    UnitPart,
    UsePackageClause,
    UseTypeClause,
    Variable,
    Variant,
    VariantPart,
    WithClause,
)
from rflx.common import file_name
from rflx.const import BUILTINS_PACKAGE, INTERNAL_PACKAGE, MAX_SCALAR_SIZE, MP_CONTEXT
from rflx.error import RecordFluxError, Severity, Subsystem, fail, warn
from rflx.identifier import ID, StrID
from rflx.integration import Integration
from rflx.model import (
    BOOLEAN,
    Composite,
    DerivedMessage,
    Enumeration,
    Field,
    Integer,
    Message,
    Model,
    Opaque,
    Refinement,
    Scalar,
    Sequence,
    Session,
    Type,
)

from . import common, const, message as message_generator
from .allocator import AllocatorGenerator
from .parser import ParserGenerator
from .serializer import SerializerGenerator
from .session import SessionGenerator

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class File:
    name: Path
    content: str


class Generator:
    def __init__(
        self,
        prefix: str = "",
        workers: int = 1,
        reproducible: bool = False,
        debug: common.Debug = common.Debug.NONE,
        ignore_unsupported_checksum: bool = False,
    ) -> None:
        self._prefix = str(ID(prefix)) if prefix else ""
        self._reproducible = reproducible
        self._debug = debug
        self._ignore_unsupported_checksum = ignore_unsupported_checksum
        self._executor = ProcessPoolExecutor(max_workers=workers, mp_context=MP_CONTEXT)
        self._template_dir = const.TEMPLATE_DIR
        assert self._template_dir.is_dir(), "template directory not found"

    def generate(
        self,
        model: Model,
        integration: Integration,
        directory: Path,
        library_files: bool = True,
        top_level_package: bool = True,
    ) -> None:
        units = self._generate(model, integration)
        self._write_files(units, directory, library_files, top_level_package)

    def _write_files(
        self,
        units: dict[ID, Unit],
        directory: Path,
        library_files: bool = True,
        top_level_package: bool = True,
    ) -> None:
        files = self._create_units(units, directory)

        if library_files:
            files.extend(self._create_library_files(directory))
        if top_level_package:
            files.extend(self._create_top_level_package(directory))

        non_updated_files = sorted(set(directory.glob("*.ad[sb]")) - {f.name for f in files})

        if non_updated_files:
            RecordFluxError(
                [
                    (
                        "partial update of generated files",
                        Subsystem.GENERATOR,
                        Severity.ERROR,
                        None,
                    ),
                    (
                        "files not generated in the current run could lead to unexpected behavior: "
                        + ", ".join(str(f.name) for f in non_updated_files),
                        Subsystem.GENERATOR,
                        Severity.INFO,
                        None,
                    ),
                    (
                        "remove the affected files or choose another directory and retry",
                        Subsystem.GENERATOR,
                        Severity.INFO,
                        None,
                    ),
                ],
            ).propagate()
        else:
            for f in files:
                log.info("Creating %s", f.name)
                f.name.write_text(f.content)

    def _create_library_files(self, directory: Path) -> list[File]:
        files = []

        for template_filename in const.LIBRARY_FILES:
            self._check_template_file(template_filename)

            prefix = f"{self._prefix}." if self._prefix else ""
            filename = Path(f"{file_name(prefix)}{template_filename}")

            template_file = (self._template_dir / template_filename).read_text()
            files.append(
                File(
                    directory / filename,
                    self._license_header
                    + "\n".join(
                        [
                            l.format(prefix=prefix)
                            for l in template_file.split("\n")
                            if "/Workarounds#" not in l
                        ],
                    ),
                ),
            )

        if self._debug == common.Debug.EXTERNAL:
            debug_package_id = self._prefix * ID("RFLX_Debug")
            files.append(
                File(
                    directory / f"{file_name(str(debug_package_id))}.ads",
                    self._license_header
                    + PackageUnit(
                        [],
                        PackageDeclaration(
                            debug_package_id,
                            [
                                SubprogramDeclaration(
                                    ProcedureSpecification(
                                        "Print",
                                        [
                                            Parameter(["Message"], "String"),
                                        ],
                                    ),
                                ),
                            ],
                            aspects=[
                                SparkMode(),
                            ],
                        ),
                        [],
                        PackageBody(debug_package_id),
                    ).ads,
                ),
            )

        return files

    def _create_top_level_package(self, directory: Path) -> list[File]:
        files = []

        if self._prefix:
            files.append(
                File(
                    Path(directory) / Path(file_name(self._prefix) + ".ads"),
                    self._license_header + f"package {self._prefix} is\n\nend {self._prefix};",
                ),
            )

        return files

    def _create_units(self, units: dict[ID, Unit], directory: Path) -> list[File]:
        files = []

        for unit in units.values():
            files.append(
                File(directory / Path(unit.name + ".ads"), self._license_header + unit.ads),
            )

            if unit.adb:
                files.append(
                    File(directory / Path(unit.name + ".adb"), self._license_header + unit.adb),
                )

        return files

    def _generate(self, model: Model, integration: Integration) -> dict[ID, Unit]:
        units: dict[ID, Unit] = {}

        for d in model.declarations:
            if d.package in [BUILTINS_PACKAGE, INTERNAL_PACKAGE]:
                continue

            log.info("Generating %s", d.identifier)

            if d.package not in units:
                unit = self._create_unit(self._prefix, d.package, terminating=False)
                units[d.package] = unit

            if isinstance(d, (Scalar, Composite)):
                units.update(self._create_type(self._prefix, d, d.package, units))

            elif isinstance(d, Message):
                # Eng/RecordFlux/RecordFlux#276
                for c in d.checksums:
                    if not self._ignore_unsupported_checksum:
                        fail(
                            "unsupported checksum (consider --ignore-unsupported-checksum option)",
                            Subsystem.GENERATOR,
                            location=c.location,
                        )
                    else:
                        warn(
                            "unsupported checksum ignored",
                            Subsystem.GENERATOR,
                            location=c.location,
                        )

                units.update(self._create_message(self._prefix, d, self._executor))

            elif isinstance(d, Refinement):
                units.update(self._create_refinement(d, units))

            elif isinstance(d, Session):
                units.update(self._create_session(d, integration))

            else:
                assert False, f'unexpected declaration "{type(d).__name__}"'

        return units

    def _create_session(self, session: Session, integration: Integration) -> dict[ID, Unit]:
        units: dict[ID, Unit] = {}
        allocator_generator = AllocatorGenerator(session.to_ir(), integration, self._prefix)

        if allocator_generator.required:
            unit = self._create_unit(
                self._prefix,
                allocator_generator.unit_identifier,
                allocator_generator.declaration_context,
                allocator_generator.body_context,
            )
            unit += allocator_generator.unit_part
            units[allocator_generator.unit_identifier] = unit

        session_generator = SessionGenerator(
            session.to_ir(),
            allocator_generator,
            self._prefix,
            debug=self._debug,
        )
        unit = self._create_unit(
            self._prefix,
            session_generator.unit_identifier,
            session_generator.declaration_context,
            session_generator.body_context,
            configuration_pragmas=[Pragma("Restrictions", [Variable("No_Streams")])],
            terminating=False,
        )
        unit += session_generator.unit_part
        units[session_generator.unit_identifier] = unit

        return units

    @staticmethod
    def _create_unit(  # noqa: PLR0913
        prefix: str,
        identifier: ID,
        declaration_context: Optional[abc.Sequence[ContextItem]] = None,
        body_context: Optional[abc.Sequence[ContextItem]] = None,
        formal_parameters: Optional[list[FormalDeclaration]] = None,
        configuration_pragmas: Optional[abc.Sequence[Pragma]] = None,
        aspects: Optional[abc.Sequence[Aspect]] = None,
        terminating: bool = True,
    ) -> PackageUnit:
        declaration_context = declaration_context if declaration_context else []
        body_context = body_context if body_context else []
        aspects = aspects if aspects else []
        configuration_pragmas = configuration_pragmas if configuration_pragmas else []

        return PackageUnit(
            [*configuration_pragmas, *const.CONFIGURATION_PRAGMAS, *declaration_context],
            PackageDeclaration(
                prefix * identifier,
                formal_parameters=formal_parameters,
                aspects=[
                    SparkMode(),
                    *([AlwaysTerminates()] if terminating else []),
                    *aspects,
                ],
            ),
            [*configuration_pragmas, *const.CONFIGURATION_PRAGMAS, *body_context],
            PackageBody(prefix * identifier, aspects=[SparkMode()]),
        )

    @staticmethod
    def _create_instantiation_unit(
        context: list[ContextItem],
        instantiation: GenericPackageInstantiation,
    ) -> InstantiationUnit:
        for p in reversed(const.CONFIGURATION_PRAGMAS):
            context.insert(0, p)

        return InstantiationUnit(context, instantiation)

    @classmethod
    def _create_message(  # noqa: PLR0912
        cls,
        prefix: str,
        message: Message,
        executor: ProcessPoolExecutor,
    ) -> dict[ID, Unit]:
        units: dict[ID, Unit] = {}

        if not message.fields:
            return units

        context: list[ContextItem] = [WithClause(prefix * const.TYPES_PACKAGE)]
        body_context: list[ContextItem] = []

        if any(t.package == BUILTINS_PACKAGE for t in message.types.values()):
            context.extend(
                [
                    WithClause(prefix * const.BUILTIN_TYPES_PACKAGE),
                    WithClause(prefix * const.BUILTIN_TYPES_CONVERSIONS_PACKAGE),
                    UsePackageClause(prefix * const.BUILTIN_TYPES_CONVERSIONS_PACKAGE),
                ],
            )

        if any(isinstance(field_type, Scalar) for field_type in message.field_types.values()):
            body_context.append(WithClause(prefix * const.TYPES_OPERATIONS_PACKAGE))

        for field_type in message.types.values():
            if field_type.package in [BUILTINS_PACKAGE, INTERNAL_PACKAGE]:
                continue

            if isinstance(field_type, Scalar) and field_type.package != message.package:
                context.extend(
                    [
                        WithClause(prefix * field_type.package),
                        UsePackageClause(prefix * field_type.package),
                    ],
                )

            elif isinstance(field_type, Sequence):
                context.append(WithClause(prefix * field_type.identifier))

        unit = cls._create_unit(prefix, message.identifier, context, body_context=body_context)
        units[message.identifier] = unit

        scalar_fields = {}
        composite_fields = []
        sequence_fields = {}
        opaque_fields = []
        fields_with_explicit_size = []
        fields_with_implicit_size = []

        for f, t in message.field_types.items():
            if isinstance(t, Scalar):
                scalar_fields[f] = t
            if isinstance(t, Composite):
                composite_fields.append(f)
                if isinstance(t, Sequence):
                    sequence_fields[f] = t
                if isinstance(t, Opaque):
                    opaque_fields.append(f)
                if any(l.has_implicit_size for l in message.incoming(f)):
                    fields_with_implicit_size.append(f)
                else:
                    fields_with_explicit_size.append(f)

        parser_generator = ParserGenerator(prefix)
        serializer_generator = SerializerGenerator(prefix)

        futures = [
            executor.submit(
                message_generator.create_use_type_clause,
                composite_fields,
                serializer_generator.requires_set_procedure(message),
            ),
            executor.submit(message_generator.create_allow_unevaluated_use_of_old),
            executor.submit(message_generator.create_field_type, message),
            executor.submit(message_generator.create_state_type),
            executor.submit(message_generator.create_cursor_type),
            executor.submit(message_generator.create_cursor_validation_functions),
            executor.submit(
                message_generator.create_cursors_invariant_function,
            ),
            executor.submit(
                message_generator.create_valid_predecessors_invariant_function,
                message,
                composite_fields,
                prefix,
            ),
            executor.submit(
                message_generator.create_valid_next_internal_function,
                message,
                composite_fields,
                prefix,
            ),
            executor.submit(
                message_generator.create_field_size_internal_function,
                message,
                prefix,
            ),
            executor.submit(
                message_generator.create_field_first_internal_function,
                message,
                prefix,
            ),
            executor.submit(
                message_generator.create_valid_context_function,
                message,
                prefix,
            ),
            executor.submit(message_generator.create_context_type, message),
            executor.submit(message_generator.create_initialize_procedure, message),
            executor.submit(
                message_generator.create_restricted_initialize_procedure,
                message,
            ),
            executor.submit(
                message_generator.create_initialized_function,
                prefix,
                message,
            ),
            executor.submit(message_generator.create_reset_procedure, prefix, message),
            executor.submit(
                message_generator.create_restricted_reset_procedure,
                prefix,
                message,
            ),
            executor.submit(
                message_generator.create_take_buffer_procedure,
                prefix,
                message,
            ),
            executor.submit(message_generator.create_copy_procedure, prefix, message),
            executor.submit(message_generator.create_read_function, prefix, message),
            executor.submit(
                message_generator.create_generic_read_procedure,
                prefix,
                message,
            ),
            executor.submit(
                message_generator.create_generic_write_procedure,
                prefix,
                message,
            ),
            executor.submit(message_generator.create_has_buffer_function),
            executor.submit(
                message_generator.create_buffer_length_function,
                prefix,
                message,
            ),
            executor.submit(message_generator.create_size_function),
            executor.submit(message_generator.create_byte_size_function),
            executor.submit(
                message_generator.create_message_last_function,
                prefix,
                message,
            ),
            executor.submit(message_generator.create_written_last_function),
            executor.submit(message_generator.create_data_procedure, prefix, message),
            executor.submit(
                message_generator.create_valid_value_function,
                prefix,
                message,
                scalar_fields,
            ),
            executor.submit(
                message_generator.create_field_condition_function,
                prefix,
                message,
            ),
            executor.submit(
                message_generator.create_field_size_function,
                prefix,
                message,
                scalar_fields,
                composite_fields,
            ),
            executor.submit(
                message_generator.create_field_first_function,
                prefix,
                message,
            ),
            executor.submit(
                message_generator.create_field_last_function,
                prefix,
                message,
                scalar_fields,
                composite_fields,
            ),
            executor.submit(
                message_generator.create_invalid_successor_function,
                message,
                SerializerGenerator.requires_set_procedure(message),
            ),
            executor.submit(message_generator.create_valid_next_function, message),
            executor.submit(
                message_generator.create_available_space_function,
                prefix,
                message,
            ),
            executor.submit(
                message_generator.create_sufficient_space_function,
                prefix,
                message,
            ),
            executor.submit(
                message_generator.create_sufficient_buffer_length_function,
                prefix,
                message,
            ),
            executor.submit(
                message_generator.create_equal_function,
                prefix,
                message,
                scalar_fields,
                composite_fields,
            ),
            executor.submit(
                message_generator.create_reset_dependent_fields_procedure,
                prefix,
                message,
            ),
            *(
                [
                    executor.submit(
                        message_generator.create_composite_field_function,
                        scalar_fields,
                        composite_fields,
                    ),
                ]
                if cls._requires_composite_field_function(
                    message,
                    scalar_fields,
                    composite_fields,
                    sequence_fields,
                )
                else []
            ),
            executor.submit(
                parser_generator.create_get_function,
                message,
                scalar_fields,
                composite_fields,
            ),
            executor.submit(
                parser_generator.create_verify_procedure,
                message,
                scalar_fields,
                composite_fields,
            ),
            executor.submit(parser_generator.create_verify_message_procedure, message),
            executor.submit(parser_generator.create_present_function),
            executor.submit(parser_generator.create_well_formed_function),
            executor.submit(parser_generator.create_valid_function),
            executor.submit(parser_generator.create_incomplete_function),
            executor.submit(parser_generator.create_invalid_function),
            executor.submit(parser_generator.create_well_formed_message_function, message),
            executor.submit(parser_generator.create_valid_message_function, message),
            executor.submit(parser_generator.create_incomplete_message_function),
            executor.submit(
                parser_generator.create_scalar_getter_functions,
                message,
                scalar_fields,
            ),
            executor.submit(
                parser_generator.create_opaque_getter_functions,
                message,
                opaque_fields,
            ),
            executor.submit(
                parser_generator.create_opaque_getter_procedures,
                message,
                opaque_fields,
            ),
            executor.submit(
                parser_generator.create_generic_opaque_getter_procedures,
                message,
                opaque_fields,
            ),
            executor.submit(serializer_generator.create_valid_size_function, message),
            executor.submit(serializer_generator.create_valid_length_function, message),
            executor.submit(
                serializer_generator.create_set_procedure,
                message,
                composite_fields,
            ),
            executor.submit(
                serializer_generator.create_scalar_setter_procedures,
                message,
                scalar_fields,
            ),
            executor.submit(
                serializer_generator.create_composite_setter_empty_procedures,
                message,
            ),
            executor.submit(
                serializer_generator.create_sequence_setter_procedures,
                message,
                sequence_fields,
            ),
            executor.submit(
                serializer_generator.create_composite_initialize_procedures,
                message,
                fields_with_explicit_size,
                fields_with_implicit_size,
            ),
            executor.submit(serializer_generator.create_opaque_setter_procedures, message),
            executor.submit(
                serializer_generator.create_generic_opaque_setter_procedures,
                message,
            ),
            executor.submit(
                message_generator.create_switch_procedures,
                prefix,
                message,
                sequence_fields,
            ),
            executor.submit(
                message_generator.create_complete_functions,
                prefix,
                message,
                sequence_fields,
            ),
            executor.submit(
                message_generator.create_update_procedures,
                prefix,
                message,
                sequence_fields,
            ),
            executor.submit(message_generator.create_cursor_function),
            executor.submit(message_generator.create_cursors_function),
            executor.submit(message_generator.create_cursors_index_function),
            executor.submit(message_generator.create_structure, prefix, message),
        ]

        for future in futures:
            unit += future.result()

        return units

    @staticmethod
    def _requires_composite_field_function(
        message: Message,
        scalar_fields: abc.Mapping[Field, Scalar],
        composite_fields: abc.Sequence[Field],
        sequence_fields: abc.Mapping[Field, Sequence],
    ) -> bool:
        return bool(
            (scalar_fields and composite_fields)
            or any(message.is_possibly_empty(f) for f in composite_fields)
            or sequence_fields,
        )

    def _create_refinement(
        self,
        refinement: Refinement,
        units: abc.Mapping[ID, Unit],
    ) -> dict[ID, Unit]:
        result: dict[ID, Unit] = {}
        identifier = refinement.package * const.REFINEMENT_PACKAGE
        null_sdu = not refinement.sdu.fields

        if identifier in units:
            unit = units[identifier]
        else:
            unit = self._create_unit(
                self._prefix,
                identifier,
                [WithClause(self._prefix * const.TYPES_PACKAGE)] if not null_sdu else [],
            )
            result[identifier] = unit

        assert isinstance(unit, PackageUnit), "unexpected unit type"

        literals = []

        def find_literals(expression: expr.Expr) -> expr.Expr:
            if isinstance(expression, expr.Literal):
                literals.append(expression)
            return expression

        refinement.condition.substituted(find_literals)

        for l in literals:
            if len(l.identifier.parts) == 2 and l.identifier.parent != refinement.package:
                unit.declaration_context.extend(
                    [
                        WithClause(self._prefix * l.identifier.parent),
                        UsePackageClause(self._prefix * l.identifier.parent),
                    ],
                )

        if not null_sdu:
            unit += UnitPart(
                [UseTypeClause(const.TYPES_INDEX), UseTypeClause(const.TYPES_BIT_INDEX)],
            )

        if refinement.pdu.package != refinement.package:
            pdu_package = (
                refinement.pdu.base.package
                if isinstance(refinement.pdu, DerivedMessage)
                else refinement.pdu.package
            )

            unit.declaration_context.extend(
                [
                    WithClause(self._prefix * pdu_package),
                    UsePackageClause(self._prefix * pdu_package),
                ],
            )

        pdu_identifier = self._prefix * refinement.pdu.identifier
        sdu_identifier = self._prefix * refinement.sdu.identifier

        unit.declaration_context.append(WithClause(pdu_identifier))

        if not null_sdu:
            unit.declaration_context.extend(
                [
                    WithClause(sdu_identifier),
                ],
            )

        condition_fields = {
            f: t
            for f, t in refinement.pdu.types.items()
            if expr.Variable(f.name) in refinement.condition
        }

        unit += self._create_contains_function(refinement, condition_fields, null_sdu)
        if not null_sdu:
            unit += UnitPart(
                [
                    UseTypeClause(f"{pdu_identifier}.Field_Cursors"),
                ],
            )
            unit += self._create_switch_procedure(refinement, condition_fields)
            unit += self._create_copy_refined_field_procedure(refinement, condition_fields)

        return result

    @classmethod
    def _create_type(
        cls,
        prefix: str,
        field_type: Type,
        message_package: ID,
        units: abc.Mapping[ID, Unit],
    ) -> dict[ID, Unit]:
        assert field_type.package != BUILTINS_PACKAGE

        result: dict[ID, Unit] = {}

        unit = units[message_package]

        assert isinstance(unit, PackageUnit)

        if isinstance(field_type, (Integer, Enumeration)):
            unit.declaration_context.append(WithClause(prefix * const.TYPES))

        if isinstance(field_type, Integer):
            unit += UnitPart(integer_types(field_type))
            unit += cls._integer_functions(prefix, field_type)
        elif isinstance(field_type, Enumeration):
            unit += UnitPart(enumeration_types(field_type))
            unit += cls._enumeration_functions(prefix, field_type)
        elif isinstance(field_type, Sequence):
            result.update(cls._create_sequence(prefix, field_type))
        else:
            assert False, f'unexpected type "{type(field_type).__name__}"'

        return result

    @classmethod
    def _create_sequence(cls, prefix: str, sequence_type: Sequence) -> dict[ID, Unit]:
        context, package = common.create_sequence_instantiation(sequence_type, prefix)
        return {
            package.identifier: cls._create_instantiation_unit(
                [
                    Pragma("SPARK_Mode"),
                    *context,
                    # Eng/RecordFlux/Workarounds#33
                    # A compiler error about a non-visible declaration of RFLX_Types inside the
                    # generic sequence package is prevented by adding a with-clause for this
                    # package.
                    Pragma(
                        "Warnings",
                        [Variable("Off"), String('unit "*RFLX_Types" is not referenced')],
                    ),
                    WithClause(prefix * const.TYPES_PACKAGE),
                    Pragma(
                        "Warnings",
                        [Variable("On"), String('unit "*RFLX_Types" is not referenced')],
                    ),
                ],
                package,
            ),
        }

    @classmethod
    def _integer_functions(cls, prefix: str, integer: Integer) -> UnitPart:
        specification: list[Declaration] = []

        constraints = (
            expr.And(
                *(
                    [expr.GreaterEqual(expr.Variable("Val"), integer.first)]
                    if integer.first.simplified() != expr.Number(0)
                    else []
                ),
                *(
                    [expr.LessEqual(expr.Variable("Val"), integer.last)]
                    if integer.last.simplified() != expr.Number(2**MAX_SCALAR_SIZE - 1)
                    else []
                ),
            ).simplified()
            if integer.first.simplified() != integer.last.simplified()
            else expr.Equal(expr.Variable("Val"), integer.first.simplified())
        )

        if constraints == expr.TRUE:
            specification.extend(
                [
                    Pragma("Warnings", [Variable("Off"), String('unused variable "Val"')]),
                    Pragma(
                        "Warnings",
                        [Variable("Off"), String('formal parameter "Val" is not referenced')],
                    ),
                ],
            )
        else:
            specification.append(UseTypeClause(prefix * const.TYPES_BASE_INT))

        specification.append(
            cls._type_validation_function(prefix, integer.name, "Val", constraints.ada_expr()),
        )

        if constraints == expr.TRUE:
            specification.extend(
                [
                    Pragma(
                        "Warnings",
                        [Variable("On"), String('formal parameter "Val" is not referenced')],
                    ),
                    Pragma("Warnings", [Variable("On"), String('unused variable "Val"')]),
                ],
            )

        specification.extend(cls._integer_conversion_functions(prefix, integer))

        return UnitPart(specification)

    @classmethod
    def _enumeration_functions(cls, prefix: str, enum: Enumeration) -> UnitPart:
        incomplete = len(enum.literals) < 2**MAX_SCALAR_SIZE

        specification: list[Declaration] = []

        validation_expression = (
            (
                Less(Variable("Val"), Pow(Number(2), enum.size.ada_expr()))
                if enum.size.simplified() != expr.Number(MAX_SCALAR_SIZE)
                else TRUE
            )
            if enum.always_valid
            else In(
                Variable("Val"),
                ChoiceList(*[value.ada_expr() for value in enum.literals.values()]),
            )
        )

        if validation_expression != TRUE:
            specification.append(UseTypeClause(prefix * const.TYPES_BASE_INT))

        specification.append(
            cls._type_validation_function(
                prefix,
                enum.name,
                "Val" if validation_expression != TRUE else "Unused_Val",
                validation_expression,
            ),
        )

        if enum.always_valid:
            specification.append(
                ExpressionFunctionDeclaration(
                    FunctionSpecification(
                        f"Valid_{enum.name}",
                        "Boolean",
                        [Parameter(["Val"], enum.name)],
                    ),
                    If(
                        [(Variable("Val.Known"), TRUE)],
                        And(
                            Call(f"Valid_{enum.name}", [Variable("Val.Raw")]),
                            NotIn(
                                Variable("Val.Raw"),
                                ChoiceList(*[value.ada_expr() for value in enum.literals.values()]),
                            ),
                        ),
                    ),
                ),
            )

        specification.append(
            ExpressionFunctionDeclaration(
                FunctionSpecification(
                    "To_Base_Integer",
                    prefix * const.TYPES_BASE_INT,
                    [
                        Parameter(
                            ["Enum"],
                            prefix
                            * (
                                common.full_enum_name(enum)
                                if enum.always_valid
                                else enum.identifier
                            ),
                        ),
                    ],
                ),
                Case(
                    Variable("Enum"),
                    [(Variable(key), value.ada_expr()) for key, value in enum.literals.items()],
                ),
            ),
        )

        conversion_function = FunctionSpecification(
            "To_Actual",
            prefix * enum.identifier,
            [Parameter(["Val"], prefix * const.TYPES_BASE_INT)],
        )
        precondition = Precondition(Call(f"Valid_{enum.name}", [Variable("Val")]))
        conversion_cases: list[tuple[Expr, Expr]] = []

        if enum.always_valid:
            specification.append(
                ExpressionFunctionDeclaration(
                    FunctionSpecification(
                        "To_Actual",
                        prefix * enum.identifier,
                        [Parameter(["Enum"], common.enum_name(enum))],
                    ),
                    Aggregate(TRUE, Variable("Enum")),
                ),
            )

            conversion_cases.extend(
                (value.ada_expr(), Aggregate(Variable("True"), Variable(key)))
                for key, value in enum.literals.items()
            )
            conversion_cases.append(
                (Variable("others"), Aggregate(Variable("False"), Variable("Val"))),
            )

            specification.append(
                ExpressionFunctionDeclaration(
                    conversion_function,
                    Case(Variable("Val"), conversion_cases),
                    [precondition],
                ),
            )

            specification.append(
                ExpressionFunctionDeclaration(
                    FunctionSpecification(
                        "To_Base_Integer",
                        prefix * const.TYPES_BASE_INT,
                        [Parameter(["Val"], prefix * enum.identifier)],
                    ),
                    If(
                        [(Variable("Val.Known"), Call("To_Base_Integer", [Variable("Val.Enum")]))],
                        Variable("Val.Raw"),
                    ),
                ),
            )
        else:
            conversion_cases.extend(
                [
                    *[(value.ada_expr(), Variable(key)) for key, value in enum.literals.items()],
                    *([(Variable("others"), Last(prefix * enum.identifier))] if incomplete else []),
                ],
            )

            specification.extend(
                [
                    Pragma("Warnings", [Variable("Off"), String("unreachable branch")]),
                    ExpressionFunctionDeclaration(
                        conversion_function,
                        Case(Variable("Val"), conversion_cases),
                        [precondition],
                    ),
                    Pragma("Warnings", [Variable("On"), String("unreachable branch")]),
                ],
            )

        return UnitPart(specification)

    def _create_contains_function(
        self,
        refinement: Refinement,
        condition_fields: abc.Mapping[Field, Type],
        null_sdu: bool,
    ) -> SubprogramUnitPart:
        pdu_identifier = self._prefix * refinement.pdu.identifier
        condition = refinement.condition
        for f, t in condition_fields.items():
            if isinstance(t, Enumeration) and t.always_valid:
                condition = expr.AndThen(
                    expr.Selected(
                        expr.Call(pdu_identifier * f"Get_{f.name}", [expr.Variable("Ctx")]),
                        "Known",
                    ),
                    condition,
                )
        condition = (
            condition.substituted(
                mapping={
                    expr.Variable(f.name): (
                        expr.Selected(
                            expr.Call(pdu_identifier * f"Get_{f.name}", [expr.Variable("Ctx")]),
                            "Enum",
                        )
                        if isinstance(t, Enumeration) and t.always_valid
                        else expr.Call(pdu_identifier * f"Get_{f.name}", [expr.Variable("Ctx")])
                    )
                    for f, t in condition_fields.items()
                },
            )
            .substituted(
                lambda e: (
                    e.copy(identifier=self._prefix * e.identifier)
                    if isinstance(e, expr.Literal) and e.identifier not in BOOLEAN.literals
                    else e
                ),
            )
            .simplified()
        )

        specification = FunctionSpecification(
            contains_function_name(refinement),
            "Boolean",
            [Parameter(["Ctx"], pdu_identifier * "Context")],
        )

        return SubprogramUnitPart(
            [
                ExpressionFunctionDeclaration(
                    specification,
                    expr.AndThen(
                        *self._refinement_conditions(refinement, "Ctx", condition_fields, null_sdu),
                        condition,
                    )
                    .simplified()
                    .ada_expr(),
                ),
            ],
        )

    def _create_switch_procedure(
        self,
        refinement: Refinement,
        condition_fields: abc.Mapping[Field, Type],
    ) -> UnitPart:
        pdu_identifier = self._prefix * refinement.pdu.identifier
        sdu_identifier = self._prefix * refinement.sdu.identifier
        pdu_context = f"{refinement.pdu.identifier.flat}_PDU_Context"
        sdu_context = f"{refinement.sdu.identifier.flat}_SDU_Context"
        refined_field_affixed_name = pdu_identifier * refinement.field.affixed_name

        specification = ProcedureSpecification(
            f"Switch_To_{refinement.field.name}",
            [
                InOutParameter([pdu_context], pdu_identifier * "Context"),
                OutParameter([sdu_context], sdu_identifier * "Context"),
            ],
        )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification,
                    [
                        Precondition(
                            And(
                                Not(Constrained(pdu_context)),
                                Not(Constrained(sdu_context)),
                                *[
                                    c.ada_expr()
                                    for c in self._refinement_conditions(
                                        refinement,
                                        pdu_context,
                                        condition_fields,
                                        null_sdu=False,
                                    )
                                ],
                                Call(
                                    self._prefix
                                    * refinement.package
                                    * const.REFINEMENT_PACKAGE
                                    * contains_function_name(refinement),
                                    [Variable(pdu_context)],
                                ),
                            ),
                        ),
                        Postcondition(
                            And(
                                Not(Call(pdu_identifier * "Has_Buffer", [Variable(pdu_context)])),
                                Call(sdu_identifier * "Has_Buffer", [Variable(sdu_context)]),
                                Equal(
                                    Selected(Variable(pdu_context), "Buffer_First"),
                                    Selected(Variable(sdu_context), "Buffer_First"),
                                ),
                                Equal(
                                    Selected(Variable(pdu_context), "Buffer_Last"),
                                    Selected(Variable(sdu_context), "Buffer_Last"),
                                ),
                                Equal(
                                    Selected(Variable(sdu_context), "First"),
                                    Call(
                                        pdu_identifier * "Field_First",
                                        [
                                            Variable(pdu_context),
                                            Variable(refined_field_affixed_name),
                                        ],
                                    ),
                                ),
                                Equal(
                                    Selected(Variable(sdu_context), "Last"),
                                    Call(
                                        pdu_identifier * "Field_Last",
                                        [
                                            Variable(pdu_context),
                                            Variable(refined_field_affixed_name),
                                        ],
                                    ),
                                ),
                                Call(sdu_identifier * "Initialized", [Variable(sdu_context)]),
                                *[
                                    Equal(e, Old(e))
                                    for e in [
                                        Selected(Variable(pdu_context), "Buffer_First"),
                                        Selected(Variable(pdu_context), "Buffer_Last"),
                                        Selected(Variable(pdu_context), "First"),
                                        Call(
                                            pdu_identifier * "Context_Cursors",
                                            [Variable(pdu_context)],
                                        ),
                                    ]
                                ],
                            ),
                        ),
                    ],
                ),
            ],
            [
                SubprogramBody(
                    specification,
                    [
                        ObjectDeclaration(
                            ["First"],
                            const.TYPES_BIT_INDEX,
                            Call(
                                pdu_identifier * "Field_First",
                                [
                                    Variable(pdu_context),
                                    Variable(refined_field_affixed_name),
                                ],
                            ),
                            constant=True,
                        ),
                        ObjectDeclaration(
                            ["Last"],
                            const.TYPES_BIT_LENGTH,
                            Call(
                                pdu_identifier * "Field_Last",
                                [
                                    Variable(pdu_context),
                                    Variable(refined_field_affixed_name),
                                ],
                            ),
                            constant=True,
                        ),
                        ObjectDeclaration(["Buffer"], const.TYPES_BYTES_PTR),
                    ],
                    [
                        CallStatement(
                            pdu_identifier * "Take_Buffer",
                            [Variable(pdu_context), Variable("Buffer")],
                        ),
                        PragmaStatement(
                            "Warnings",
                            [Variable("Off"), String('unused assignment to "Buffer"')],
                        ),
                        CallStatement(
                            sdu_identifier * "Initialize",
                            [
                                Variable(sdu_context),
                                Variable("Buffer"),
                                Variable("First"),
                                Variable("Last"),
                                Variable("Last"),
                            ],
                        ),
                        PragmaStatement(
                            "Warnings",
                            [Variable("On"), String('unused assignment to "Buffer"')],
                        ),
                    ],
                ),
            ],
        )

    def _create_copy_refined_field_procedure(
        self,
        refinement: Refinement,
        condition_fields: abc.Mapping[Field, Type],
    ) -> UnitPart:
        pdu_identifier = self._prefix * refinement.pdu.identifier
        sdu_identifier = self._prefix * refinement.sdu.identifier
        pdu_context = ID(refinement.pdu.identifier.flat + "_PDU_Context")
        sdu_context = ID(refinement.sdu.identifier.flat + "_SDU_Context")
        refined_field_affixed_name = pdu_identifier * refinement.field.affixed_name

        specification = ProcedureSpecification(
            f"Copy_{refinement.field.name}",
            [
                Parameter([pdu_context], pdu_identifier * "Context"),
                InOutParameter([sdu_context], sdu_identifier * "Context"),
            ],
        )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification,
                    [
                        Precondition(
                            AndThen(
                                Not(Constrained(sdu_context)),
                                Call(sdu_identifier * "Has_Buffer", [Variable(sdu_context)]),
                                *[
                                    c.ada_expr()
                                    for c in self._refinement_conditions(
                                        refinement,
                                        pdu_context,
                                        condition_fields,
                                        null_sdu=False,
                                    )
                                ],
                                Call(
                                    self._prefix
                                    * refinement.package
                                    * const.REFINEMENT_PACKAGE
                                    * contains_function_name(refinement),
                                    [Variable(pdu_context)],
                                ),
                                GreaterEqual(
                                    Add(
                                        Call(
                                            const.TYPES_TO_LAST_BIT_INDEX,
                                            [Variable(sdu_context * "Buffer_Last")],
                                        ),
                                        -Call(
                                            const.TYPES_TO_FIRST_BIT_INDEX,
                                            [Variable(sdu_context * "Buffer_First")],
                                        ),
                                        Number(1),
                                    ),
                                    Call(
                                        pdu_identifier * "Field_Size",
                                        [
                                            Variable(pdu_context),
                                            Variable(refined_field_affixed_name),
                                        ],
                                    ),
                                ),
                                Less(
                                    Add(
                                        Call(
                                            const.TYPES_TO_FIRST_BIT_INDEX,
                                            [Variable(sdu_context * "Buffer_First")],
                                        ),
                                        Call(
                                            pdu_identifier * "Field_Size",
                                            [
                                                Variable(pdu_context),
                                                Variable(refined_field_affixed_name),
                                            ],
                                        ),
                                        -Number(1),
                                    ),
                                    Last(const.TYPES_BIT_INDEX),
                                ),
                            ),
                        ),
                        Postcondition(
                            And(
                                Call(pdu_identifier * "Has_Buffer", [Variable(pdu_context)]),
                                Call(sdu_identifier * "Has_Buffer", [Variable(sdu_context)]),
                                Call(sdu_identifier * "Initialized", [Variable(sdu_context)]),
                                *[
                                    Equal(e, Old(e))
                                    for e in [
                                        Selected(Variable(sdu_context), "Buffer_First"),
                                        Selected(Variable(sdu_context), "Buffer_Last"),
                                    ]
                                ],
                            ),
                        ),
                    ],
                ),
            ],
            [
                SubprogramBody(
                    specification,
                    [
                        ObjectDeclaration(
                            ["First"],
                            const.TYPES_BIT_INDEX,
                            Call(
                                const.TYPES_TO_FIRST_BIT_INDEX,
                                [Variable(sdu_context * "Buffer_First")],
                            ),
                            constant=True,
                        ),
                        ObjectDeclaration(
                            ["Size"],
                            const.TYPES_BIT_INDEX,
                            Call(
                                pdu_identifier * "Field_Size",
                                [
                                    Variable(pdu_context),
                                    Variable(refined_field_affixed_name),
                                ],
                            ),
                            constant=True,
                        ),
                        ObjectDeclaration(["Buffer"], const.TYPES_BYTES_PTR),
                    ],
                    [
                        PragmaStatement(
                            "Warnings",
                            [
                                Variable("Off"),
                                String(
                                    f'"{sdu_context.ada_str}" is set by "Take_Buffer"'
                                    " but not used after the call",
                                ),
                            ],
                        ),
                        CallStatement(
                            sdu_identifier * "Take_Buffer",
                            [Variable(sdu_context), Variable("Buffer")],
                        ),
                        PragmaStatement(
                            "Warnings",
                            [
                                Variable("On"),
                                String(
                                    f'"{sdu_context.ada_str}" is set by "Take_Buffer"'
                                    " but not used after the call",
                                ),
                            ],
                        ),
                        CallStatement(
                            pdu_identifier * f"Get_{refinement.field.name}",
                            [
                                Variable(pdu_context),
                                Slice(
                                    Variable("Buffer.all"),
                                    First("Buffer"),
                                    Add(
                                        First("Buffer"),
                                        Call(
                                            const.TYPES_INDEX,
                                            [
                                                Call(
                                                    const.TYPES_TO_LENGTH,
                                                    [Variable("Size")],
                                                ),
                                            ],
                                        ),
                                        -Number(1),
                                    ),
                                ),
                            ],
                        ),
                        CallStatement(
                            sdu_identifier * "Initialize",
                            [
                                Variable(sdu_context),
                                Variable("Buffer"),
                                Variable("First"),
                                Add(
                                    Variable("First"),
                                    Variable("Size"),
                                    -Number(1),
                                ),
                                Add(
                                    Variable("First"),
                                    Variable("Size"),
                                    -Number(1),
                                ),
                            ],
                        ),
                    ],
                ),
            ],
        )

    def _check_template_file(self, filename: str) -> None:
        assert self._template_dir.joinpath(
            filename,
        ).is_file(), f'template file not found: "{filename}"'

    @cached_property
    def _license_header(self) -> str:
        if self._reproducible:
            return (
                common.comment_box(
                    [
                        "",
                        "Generated by RecordFlux",
                        "",
                        "Copyright (C) AdaCore GmbH",
                        "",
                        "SPDX-License-Identifier: Apache-2.0",
                        "",
                    ],
                )
                + "\n"
            )

        today = date.today()  # noqa: DTZ011
        return (
            common.comment_box(
                [
                    "",
                    f"Generated by RecordFlux {__version__} on {today}",
                    "",
                    f"Copyright (C) 2018-{today.year} AdaCore GmbH",
                    "",
                    "SPDX-License-Identifier: Apache-2.0",
                    "",
                ],
            )
            + "\n"
        )

    @staticmethod
    def _type_validation_function(
        prefix: str,
        type_name: str,
        enum_value: str,
        validation_expression: Expr,
    ) -> Subprogram:
        return ExpressionFunctionDeclaration(
            FunctionSpecification(
                f"Valid_{type_name}",
                "Boolean",
                [Parameter([enum_value], prefix * const.TYPES_BASE_INT)],
            ),
            validation_expression,
        )

    @staticmethod
    def _integer_conversion_functions(prefix: str, integer: Integer) -> list[Subprogram]:
        return [
            ExpressionFunctionDeclaration(
                FunctionSpecification(
                    "To_Base_Integer",
                    prefix * const.TYPES_BASE_INT,
                    [Parameter(["Val"], prefix * integer.identifier)],
                ),
                Call(prefix * const.TYPES_BASE_INT, [Variable("Val")]),
            ),
            ExpressionFunctionDeclaration(
                FunctionSpecification(
                    "To_Actual",
                    prefix * integer.identifier,
                    [Parameter(["Val"], prefix * const.TYPES_BASE_INT)],
                ),
                Call(prefix * integer.identifier, [Variable("Val")]),
                [Precondition(Call(f"Valid_{integer.name}", [Variable("Val")]))],
            ),
        ]

    def _refinement_conditions(
        self,
        refinement: Refinement,
        pdu_context: StrID,
        condition_fields: abc.Mapping[Field, Type],
        null_sdu: bool,
    ) -> list[expr.Expr]:
        pdu_identifier = self._prefix * refinement.pdu.identifier

        conditions: list[expr.Expr] = [
            expr.Call(pdu_identifier * "Has_Buffer", [expr.Variable(pdu_context)]),
        ]

        if null_sdu:
            conditions.extend(
                [
                    expr.Call(
                        pdu_identifier * "Well_Formed",
                        [
                            expr.Variable(pdu_context),
                            expr.Variable(pdu_identifier * refinement.field.affixed_name),
                        ],
                    ),
                    expr.Not(
                        expr.Call(
                            pdu_identifier * "Present",
                            [
                                expr.Variable(pdu_context),
                                expr.Variable(pdu_identifier * refinement.field.affixed_name),
                            ],
                        ),
                    ),
                ],
            )
        else:
            conditions.append(
                expr.Call(
                    pdu_identifier * "Present",
                    [
                        expr.Variable(pdu_context),
                        expr.Variable(pdu_identifier * refinement.field.affixed_name),
                    ],
                ),
            )

        conditions.extend(
            [
                expr.Call(
                    pdu_identifier * "Valid",
                    [
                        expr.Variable(pdu_context),
                        expr.Variable(pdu_identifier * f.affixed_name),
                    ],
                )
                for f in condition_fields
            ],
        )

        return conditions


def integer_types(integer: Integer) -> list[Declaration]:
    return [
        RangeType(
            integer.name,
            integer.first_expr.ada_expr(),
            integer.last_expr.ada_expr(),
            aspects=[SizeAspect(integer.size_expr.ada_expr())],
        ),
    ]


def enumeration_types(enum: Enumeration) -> list[Declaration]:
    types: list[Declaration] = []

    types.append(
        EnumerationType(
            common.enum_name(enum) if enum.always_valid else enum.name,
            {k: Number(v.value) for k, v in enum.literals.items()},
            enum.size_expr.ada_expr(),
        ),
    )
    if enum.always_valid:
        types.append(
            RecordType(
                enum.name,
                [],
                [Discriminant(["Known"], "Boolean", FALSE)],
                VariantPart(
                    "Known",
                    [
                        Variant([TRUE], [Component("Enum", common.enum_name(enum))]),
                        Variant([FALSE], [Component("Raw", const.TYPES_BASE_INT)]),
                    ],
                ),
            ),
        )

    return types


def contains_function_name(refinement: Refinement) -> str:
    return common.contains_function_name(
        refinement.package,
        refinement.pdu.identifier,
        refinement.sdu.identifier,
        refinement.field.identifier,
    )
